import pharmpy.results
import pharmpy.tools.iiv.algorithms as algorithms
import pharmpy.tools.modelsearch.tool
from pharmpy.modeling import (
    add_pk_iiv,
    copy_model,
    create_joint_distribution,
    summarize_modelfit_results,
    update_inits,
)
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow


def create_workflow(
    algorithm,
    add_iivs=False,
    iiv_as_fullblock=False,
    rankfunc='bic',
    cutoff=None,
    model=None,
):
    algorithm_func = getattr(algorithms, algorithm)

    # FIXME: must currently be a model, cannot be a task
    if add_iivs:
        model_iiv = copy_model(model, f'{model.name}_add_iiv')
        _add_iiv(iiv_as_fullblock, model_iiv)
        iivs = model_iiv.random_variables.iiv
    else:
        iivs = model.random_variables.iiv

    wf = Workflow()
    wf.name = 'iiv'

    start_task = Task('start_iiv', start, add_iivs, iiv_as_fullblock, model)
    wf.add_task(start_task)

    if not model.modelfit_results or add_iivs:
        wf_fit = create_fit_workflow(n=1)
        wf.insert_workflow(wf_fit)
        start_model_task = wf_fit.output_tasks
    else:
        start_model_task = [start_task]

    task_update_inits = Task('update_inits_start_model', _update_inits_start_model)
    wf.add_task(task_update_inits, predecessors=wf.output_tasks)

    wf_method, model_features = algorithm_func(iivs)
    wf.insert_workflow(wf_method)

    task_result = Task(
        'results',
        post_process_results,
        rankfunc,
        cutoff,
        model_features,
    )

    wf.add_task(task_result, predecessors=start_model_task + wf.output_tasks)

    return wf


def start(add_iivs, iiv_as_fullblock, model):
    if add_iivs:
        model = copy_model(model, f'{model.name}_add_iiv')
        _add_iiv(iiv_as_fullblock, model)
    return model


def _add_iiv(iiv_as_fullblock, model):
    add_pk_iiv(model)
    if iiv_as_fullblock:
        create_joint_distribution(model)
    return model


def _update_inits_start_model(model):
    try:
        update_inits(model)
    except ValueError:
        pass
    return model


def post_process_results(rankfunc, cutoff, model_features, *models):
    start_model, res_models = models

    if isinstance(res_models, tuple):
        res_models = list(res_models)
    else:
        res_models = [res_models]

    summary_tool = pharmpy.tools.modelsearch.tool.create_summary(
        res_models,
        start_model,
        rankfunc,
        cutoff,
        model_features,
        rank_by_not_worse=True,
        bic_type='iiv',
    )

    best_model_name = summary_tool['rank'].idxmin()
    try:
        best_model = [model for model in res_models if model.name == best_model_name][0]
    except IndexError:
        best_model = start_model

    summary_models = summarize_modelfit_results([start_model] + res_models)

    res = IIVResults(
        summary_tool=summary_tool,
        summary_models=summary_models,
        best_model=best_model,
        start_model=start_model,
        models=res_models,
    )

    return res


class IIVResults(pharmpy.results.Results):
    def __init__(
        self, summary_tool=None, summary_models=None, best_model=None, start_model=None, models=None
    ):
        self.summary_tool = summary_tool
        self.summary_models = summary_models
        self.best_model = best_model
        self.start_model = start_model
        self.models = models



import copy
from collections.abc import MutableSequence

import pandas as pd


class EstimationSteps(MutableSequence):
    """A sequence of estimation steps

    Parameters
    ----------
    steps : EstimationStepsi, iterable or None
        Used for initialization
    """

    def __init__(self, steps=None):
        if isinstance(steps, EstimationSteps):
            self._steps = copy.deepcopy(steps._steps)
        elif steps is None:
            self._steps = []
        else:
            self._steps = list(steps)

    def __getitem__(self, i):
        return self._steps[i]

    def __setitem__(self, i, value):
        self._steps[i] = value

    def __delitem__(self, i):
        del self._steps[i]

    def __len__(self):
        return len(self._steps)

    def __eq__(self, other):
        if len(self) != len(other):
            return False
        for s1, s2 in zip(self, other):
            if s1 != s2:
                return False
        return True

    def insert(self, i, value):
        self._steps.insert(i, value)

    def copy(self):
        """Create a deepcopy"""
        return copy.deepcopy(self)

    def to_dataframe(self):
        """Convert to DataFrame

        Use this to create an overview of all estimation steps

        Results
        -------
        pd.DataFrame
            DataFrame overview

        >>> from pharmpy.modeling import load_example_model
        >>> model = load_example_model("pheno")
        >>> model.estimation_steps.to_dataframe()   # doctest: +ELLIPSIS
          method  interaction   cov  ...  auto keep_every_nth_iter  tool_options
        0   FOCE         True  True  ...  None                None            {}
        """
        method = [s.method for s in self._steps]
        interaction = [s.interaction for s in self._steps]
        cov = [s.cov for s in self._steps]
        evaluation = [s.evaluation for s in self._steps]
        maximum_evaluations = [s.maximum_evaluations for s in self._steps]
        laplace = [s.laplace for s in self._steps]
        isample = [s.isample for s in self._steps]
        niter = [s.niter for s in self._steps]
        auto = [s.auto for s in self._steps]
        keep_every_nth_iter = [s.keep_every_nth_iter for s in self._steps]
        tool_options = [s.tool_options for s in self._steps]
        df = pd.DataFrame(
            {
                'method': method,
                'interaction': interaction,
                'cov': cov,
                'evaluation': evaluation,
                'maximum_evaluations': maximum_evaluations,
                'laplace': laplace,
                'isample': isample,
                'niter': niter,
                'auto': auto,
                'keep_every_nth_iter': keep_every_nth_iter,
                'tool_options': tool_options,
            }
        )
        return df

    def __repr__(self):
        if len(self) == 0:
            return "EstimationSteps()"
        return self.to_dataframe().to_string()

    def _repr_html_(self):
        if len(self) == 0:
            return "EstimationSteps()"
        else:
            return self.to_dataframe().to_html()


class EstimationStep:
    """Definition of one estimation operation"""

    """Supported estimation methods
    """
    supported_methods = ['FO', 'FOCE', 'ITS', 'IMPMAP', 'IMP', 'SAEM', 'BAYES']

    def __init__(
        self,
        method,
        interaction=False,
        cov=False,
        evaluation=False,
        maximum_evaluations=None,
        laplace=False,
        isample=None,
        niter=None,
        auto=None,
        keep_every_nth_iter=None,
        residuals=None,
        predictions=None,
        tool_options=None,
    ):
        method = self._canonicalize_and_check_method(method)
        self.method = method
        self.interaction = interaction
        self.cov = cov
        self.evaluation = evaluation
        self.maximum_evaluations = maximum_evaluations
        self.laplace = laplace
        self.isample = isample
        self.niter = niter
        self.auto = auto
        self.keep_every_nth_iter = keep_every_nth_iter
        if residuals is None:
            self.residuals = []
        else:
            self.residuals = residuals
        if predictions is None:
            self.predictions = []
        else:
            self.predictions = predictions
        if tool_options is None:
            self.tool_options = dict()
        else:
            self.tool_options = tool_options

    def _canonicalize_and_check_method(self, method):
        method = method.upper()
        if method not in self.supported_methods:
            raise ValueError(
                f'EstimationStep: {method} not recognized. Use any of {self.supported_methods}.'
            )
        return method

    @property
    def method(self):
        """Name of the estimation method"""
        return self._method

    @method.setter
    def method(self, value):
        method = self._canonicalize_and_check_method(value)
        self._method = method

    @property
    def maximum_evaluations(self):
        """Maximum allowable number of evaluations of the objective function"""
        return self._maximum_evaluations

    @maximum_evaluations.setter
    def maximum_evaluations(self, value):
        if value is not None and value < 1:
            raise ValueError(
                'Number of maximum evaluations must be more than one, use '
                'evaluation=True or tool_options for special cases (e.g. 0 and -1'
                'in NONMEM)'
            )
        self._maximum_evaluations = value

    @property
    def interaction(self):
        """Preserve eta-epsilon interaction in the computation of the objective function"""
        return self._interaction

    @interaction.setter
    def interaction(self, value):
        self._interaction = bool(value)

    @property
    def evaluation(self):
        """Only perform model evaluation"""
        return self._evaluation

    @evaluation.setter
    def evaluation(self, value):
        self._evaluation = bool(value)

    @property
    def cov(self):
        """Should the parameter uncertainty be estimated?"""
        return self._cov

    @cov.setter
    def cov(self, value):
        self._cov = bool(value)

    @property
    def laplace(self):
        """Use the laplacian method"""
        return self._laplace

    @laplace.setter
    def laplace(self, value):
        self._laplace = bool(value)

    @property
    def isample(self):
        """Number of samples per subject (or similar) for EM methods"""
        return self._isample

    @isample.setter
    def isample(self, value):
        self._isample = value

    @property
    def niter(self):
        """Number of iterations for EM methods"""
        return self._niter

    @niter.setter
    def niter(self, value):
        self._niter = value

    @property
    def auto(self):
        """Let estimation tool automatically add settings"""
        return self._auto

    @auto.setter
    def auto(self, value):
        self._auto = value

    @property
    def keep_every_nth_iter(self):
        """Keep results for every nth iteration"""
        return self._keep_every_nth_iter

    @keep_every_nth_iter.setter
    def keep_every_nth_iter(self, value):
        self._keep_every_nth_iter = value

    @property
    def residuals(self):
        """List of residuals to calculate"""
        return self._residuals

    @residuals.setter
    def residuals(self, value):
        self._residuals = value

    @property
    def predictions(self):
        """List of predictions to estimate"""
        return self._predictions

    @predictions.setter
    def predictions(self, value):
        self._predictions = value

    @property
    def tool_options(self):
        """Dictionary of tool specific options"""
        return self._tool_options

    @tool_options.setter
    def tool_options(self, value):
        self._tool_options = value

    def __eq__(self, other):
        return (
            self.method == other.method
            and self.interaction == other.interaction
            and self.cov == other.cov
            and self.evaluation == other.evaluation
            and self.maximum_evaluations == other.maximum_evaluations
            and self.laplace == other.laplace
            and self.isample == other.isample
            and self.niter == other.niter
            and self.auto == other.auto
            and self.keep_every_nth_iter == other.keep_every_nth_iter
            and self.tool_options == other.tool_options
        )

    def __repr__(self):
        return (
            f'EstimationStep("{self.method}", interaction={self.interaction}, '
            f'cov={self.cov}, evaluation={self.evaluation}, '
            f'maximum_evaluations={self.maximum_evaluations}, laplace={self.laplace}, '
            f'isample={self.isample}, niter={self.niter}, auto={self.auto}, '
            f'keep_every_nth_iter={self.keep_every_nth_iter}, '
            f'tool_options={self.tool_options})'
        )

    def copy(self):
        """Create a deep copy"""
        return copy.deepcopy(self)



from pathlib import Path

import pytest

import pharmpy.data
import pharmpy.model
import pharmpy.symbols
from pharmpy import Model
from pharmpy.modeling import convert_model, create_symbol
from pharmpy.plugins.nonmem.dataset import read_nonmem_dataset

tabpath = Path(__file__).parent / 'testdata' / 'nonmem' / 'pheno_real_linbase.tab'
lincorrect = read_nonmem_dataset(
    tabpath,
    ignore_character='@',
    colnames=['ID', 'G11', 'G21', 'H11', 'CIPREDI', 'DV', 'PRED', 'RES', 'WRES'],
)


@pytest.mark.parametrize(
    'stem,force_numbering,symbol_name', [('DV', False, 'DV1'), ('X', False, 'X'), ('X', True, 'X1')]
)
def test_create_symbol(testdata, stem, force_numbering, symbol_name):
    model = Model.create_model(testdata / 'nonmem' / 'pheno_real.mod')
    symbol = create_symbol(model, stem=stem, force_numbering=force_numbering)

    assert symbol.name == symbol_name


def test_to_generic_model(testdata):
    path = testdata / 'nonmem' / 'pheno.mod'
    nm_model = Model.create_model(path)
    model = convert_model(nm_model, 'generic')

    assert id(model.parameters) != id(nm_model.parameters)
    assert model.parameters == nm_model.parameters
    assert id(model.random_variables) != id(nm_model.random_variables)
    assert model.random_variables == nm_model.random_variables
    assert model.name == nm_model.name
    assert id(model.statements) != id(nm_model.statements)
    assert model.statements == nm_model.statements
    assert type(model) == pharmpy.model.Model



import functools

import pytest

from pharmpy.modeling import (
    set_peripheral_compartments,
    set_transit_compartments,
    set_zero_order_absorption,
)
from pharmpy.tools.modelsearch.algorithms import (
    _is_allowed,
    exhaustive,
    exhaustive_stepwise,
    reduced_stepwise,
)
from pharmpy.tools.modelsearch.mfl import ModelFeatures


def test_exhaustive_algorithm():
    mfl = 'ABSORPTION(ZO);PERIPHERALS(1)'
    wf, _, model_features = exhaustive(mfl, False, False, False)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == 3
    assert list(model_features.values())[-1] == ('ABSORPTION(ZO)', 'PERIPHERALS(1)')


@pytest.mark.parametrize(
    'mfl, add_iivs, no_of_models, last_model_features',
    [
        (
            'ABSORPTION(ZO)\nPERIPHERALS(1)',
            False,
            4,
            ('PERIPHERALS(1)', 'ABSORPTION(ZO)'),
        ),
        ('ABSORPTION(ZO);TRANSITS(1)', False, 2, ('TRANSITS(1)',)),
        (
            'ABSORPTION([ZO,SEQ-ZO-FO]);PERIPHERALS(1)',
            False,
            7,
            ('PERIPHERALS(1)', 'ABSORPTION(ZO)'),
        ),
        (
            'ABSORPTION(ZO);PERIPHERALS([1, 2])',
            False,
            8,
            ('PERIPHERALS(1)', 'PERIPHERALS(2)', 'ABSORPTION(ZO)'),
        ),
        (
            'ABSORPTION(SEQ-ZO-FO);LAGTIME()',
            False,
            2,
            ('LAGTIME()',),
        ),
        (
            'ABSORPTION(ZO);LAGTIME();PERIPHERALS(1)',
            False,
            15,
            ('LAGTIME()', 'PERIPHERALS(1)', 'ABSORPTION(ZO)'),
        ),
        (
            'ABSORPTION(ZO);LAGTIME();PERIPHERALS([1,2]);ELIMINATION(ZO)',
            False,
            170,
            ('LAGTIME()', 'PERIPHERALS(1)', 'PERIPHERALS(2)', 'ELIMINATION(ZO)', 'ABSORPTION(ZO)'),
        ),
        (
            'LAGTIME();TRANSITS(1);PERIPHERALS(1)',
            True,
            7,
            ('LAGTIME()', 'PERIPHERALS(1)'),
        ),
        (
            'ABSORPTION([ZO,SEQ-ZO-FO]);ELIMINATION(MM)',
            False,
            7,
            ('ELIMINATION(MM)', 'ABSORPTION(ZO)'),
        ),
        (
            'ABSORPTION([ZO,SEQ-ZO-FO]);LAGTIME();TRANSITS([1,3,10],*);'
            'PERIPHERALS(1);ELIMINATION([MM,MIX-FO-MM])',
            False,
            246,
            ('LAGTIME()', 'PERIPHERALS(1)', 'ELIMINATION(MM)', 'ABSORPTION(ZO)'),
        ),
    ],
)
def test_exhaustive_stepwise_algorithm(mfl, add_iivs, no_of_models, last_model_features):
    wf, _, model_features = exhaustive_stepwise(mfl, add_iivs, False, False)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == no_of_models
    assert list(model_features.values())[-1] == last_model_features


@pytest.mark.parametrize(
    'mfl, no_of_models, last_model_features',
    [
        (
            'ABSORPTION(ZO);LAGTIME();PERIPHERALS(1)',
            12,
            'ABSORPTION(ZO)',
        ),
        (
            'ABSORPTION(ZO);LAGTIME();PERIPHERALS([1,2]);ELIMINATION(ZO)',
            52,
            'PERIPHERALS(2)',
        ),
        ('ABSORPTION([ZO,SEQ-ZO-FO]);ELIMINATION(MM)', 7, 'ABSORPTION(ZO)'),
        (
            'ABSORPTION([ZO,SEQ-ZO-FO]);LAGTIME();TRANSITS([1,3,10],*);'
            'PERIPHERALS(1);ELIMINATION([MM,MIX-FO-MM])',
            159,
            'ABSORPTION(ZO)',
        ),
    ],
)
def test_reduced_stepwise_algorithm(mfl, no_of_models, last_model_features):
    wf, _, model_features = reduced_stepwise(mfl, False, False, False)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == no_of_models
    assert all(task.name == 'run0' for task in wf.output_tasks)
    assert list(model_features.values())[-1] == last_model_features


def test_is_allowed():
    features = ModelFeatures('ABSORPTION(ZO);PERIPHERALS(1)')
    feat_previous = []
    feat_current, func_current = 'ABSORPTION(ZO)', set_zero_order_absorption
    assert _is_allowed(feat_current, func_current, feat_previous, features)
    assert _is_allowed(feat_current, func_current, feat_current, features) is False

    features = ModelFeatures('ABSORPTION([ZO,SEQ-ZO-FO])')
    feat_previous = ['ABSORPTION(SEQ-ZO-FO)']
    feat_current, func_current = 'ABSORPTION(ZO)', set_zero_order_absorption
    assert _is_allowed(feat_current, func_current, feat_previous, features) is False

    features = ModelFeatures('PERIPHERALS([1,2])')
    feat_previous = []
    feat_current, func_current = 'PERIPHERALS(1)', functools.partial(
        set_peripheral_compartments, n=1
    )
    assert _is_allowed(feat_current, func_current, feat_previous, features)
    feat_previous = [feat_current]
    feat_current, func_current = 'PERIPHERALS(2)', functools.partial(
        set_peripheral_compartments, n=2
    )
    assert _is_allowed(feat_current, func_current, feat_previous, features)

    features = ModelFeatures('PERIPHERALS([1,2])')
    feat_previous = ['PERIPHERALS(1)']
    feat_current, func_current = 'PERIPHERALS(1)', functools.partial(
        set_peripheral_compartments, n=1
    )
    assert _is_allowed(feat_current, func_current, feat_previous, features) is False

    features = ModelFeatures('PERIPHERALS([1,2])')
    feat_previous = []
    feat_current, func_current = 'PERIPHERALS(2)', functools.partial(
        set_peripheral_compartments, n=2
    )
    assert _is_allowed(feat_current, func_current, feat_previous, features) is False

    features = ModelFeatures('PERIPHERALS(2)')
    feat_previous = []
    feat_current, func_current = 'PERIPHERALS(2)', functools.partial(
        set_peripheral_compartments, n=2
    )
    assert _is_allowed(feat_current, func_current, feat_previous, features)


@pytest.mark.parametrize(
    'code,args',
    [
        ('ABSORPTION(FO)', {'FO'}),
        ('ABSORPTION(* )', {'FO', 'ZO', 'SEQ-ZO-FO'}),
        ('ABSORPTION([ZO,FO])', {'FO', 'ZO'}),
        ('ABSORPTION([ZO,  FO])', {'FO', 'ZO'}),
        ('ABSORPTION( [   SEQ-ZO-FO,  FO   ]  )', {'SEQ-ZO-FO', 'FO'}),
        ('absorption([zo, fo])', {'FO', 'ZO'}),
        ('ABSORPTION(FO);ABSORPTION(ZO)', {'FO', 'ZO'}),
        ('ABSORPTION(FO)\nABSORPTION([FO, SEQ-ZO-FO])', {'FO', 'SEQ-ZO-FO'}),
    ],
)
def test_mfl_absorption(code, args):
    mfl = ModelFeatures(code)
    assert mfl.absorption.args == args


@pytest.mark.parametrize(
    'code,args',
    [
        ('ELIMINATION(FO)', {'FO'}),
        ('ELIMINATION( *)', {'FO', 'ZO', 'MM', 'MIX-FO-MM'}),
        ('ELIMINATION([ZO,FO])', {'FO', 'ZO'}),
        ('ELIMINATION([ZO,  FO])', {'FO', 'ZO'}),
        ('ELIMINATION( [   MIX-FO-MM,  FO   ]  )', {'MIX-FO-MM', 'FO'}),
        ('elimination([zo, fo])', {'FO', 'ZO'}),
        ('ELIMINATION(FO);ABSORPTION(ZO)', {'FO'}),
    ],
)
def test_mfl_elimination(code, args):
    mfl = ModelFeatures(code)
    assert mfl.elimination.args == args


@pytest.mark.parametrize(
    'code,args,depot',
    [
        ('TRANSITS(0)', {0}, 'DEPOT'),
        ('TRANSITS([0, 1])', {0, 1}, 'DEPOT'),
        ('TRANSITS([0, 2, 4])', {0, 2, 4}, 'DEPOT'),
        ('TRANSITS(0..1)', {0, 1}, 'DEPOT'),
        ('TRANSITS(1..4)', {1, 2, 3, 4}, 'DEPOT'),
        ('TRANSITS(1..4); TRANSITS(5)', {1, 2, 3, 4, 5}, 'DEPOT'),
        ('TRANSITS(0);PERIPHERALS(0)', {0}, 'DEPOT'),
        ('TRANSITS(1..4, DEPOT)', {1, 2, 3, 4}, 'DEPOT'),
        ('TRANSITS(1..4, NODEPOT)', {1, 2, 3, 4}, 'NODEPOT'),
        ('TRANSITS(1..4, *)', {1, 2, 3, 4}, '*'),
    ],
)
def test_mfl_transits(code, args, depot):
    mfl = ModelFeatures(code)
    assert mfl.transits.args == args
    assert mfl.transits.depot == depot


def test_mfl_transits_depot():
    mfl = ModelFeatures('TRANSITS(1, *)')
    func_depot = functools.partial(set_transit_compartments, n=1)
    assert mfl.transits._funcs['TRANSITS(1)'].keywords == func_depot.keywords
    func_nodepot = functools.partial(set_transit_compartments, n=2, keep_depot=False)
    assert mfl.transits._funcs['TRANSITS(1, NODEPOT)'].keywords == func_nodepot.keywords


@pytest.mark.parametrize(
    'code,args',
    [
        ('PERIPHERALS(0)', {0}),
        ('PERIPHERALS([0, 1])', {0, 1}),
        ('PERIPHERALS([0, 2, 4])', {0, 2, 4}),
        ('PERIPHERALS(0..1)', {0, 1}),
        ('PERIPHERALS(1..4)', {1, 2, 3, 4}),
        ('PERIPHERALS(1..4); PERIPHERALS(5)', {1, 2, 3, 4, 5}),
    ],
)
def test_mfl_peripherals(code, args):
    mfl = ModelFeatures(code)
    assert mfl.peripherals.args == args


@pytest.mark.parametrize(
    'code,args',
    [
        ('LAGTIME()', None),
    ],
)
def test_mfl_lagtime(code, args):
    mfl = ModelFeatures(code)
    assert mfl.lagtime.args == args


@pytest.mark.parametrize(
    'code',
    [
        ('ABSORPTION(ILLEGAL)'),
        ('ELIMINATION(ALSOILLEGAL)'),
        ('LAGTIME(0)'),
        ('TRANSITS(*)'),
    ],
)
def test_illegal_mfl(code):
    with pytest.raises(Exception):
        ModelFeatures(code)



import sympy

from pharmpy.statements import Assignment, ODESystem


def _extract_minus(expr):
    if expr.is_Mul and expr.args[0] == -1:
        return sympy.Mul(*expr.args[1:])
    else:
        return expr


def get_unit_of(model, variable):
    """Derive the physical unit of a variable in the model

    Unit information for the dataset needs to be available.
    The variable can be defined in the code, a dataset olumn, a parameter
    or a random variable.

    Parameters
    ----------
    model : Model
        Pharmpy model object
    variable : str or Symbol
        Find physical unit of this variable

    Returns
    -------
    unit expression
        A sympy physics.units expression

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, get_unit_of
    >>> model = load_example_model("pheno")
    >>> get_unit_of(model, "Y")
    milligram/liter
    >>> get_unit_of(model, "V")
    liter
    >>> get_unit_of(model, "WGT")
    kilogram
    """
    if isinstance(variable, str):
        symbol = sympy.Symbol(variable)
    else:
        symbol = variable
        variable = variable.name

    di = model.datainfo
    if variable in di.names:
        return di[variable].unit

    y = model.dependent_variable
    input_units = {sympy.Symbol(col.name): col.unit for col in di}
    input_units[sympy.exp] = 1
    unit_eqs = []
    unit_eqs.append(y - di[di.dv_column.name].unit)

    for s in model.statements:
        if isinstance(s, Assignment):
            expr = sympy.expand(s.expression.subs(input_units))
            if expr.is_Add:
                for term in expr.args:
                    unit_eqs.append(s.symbol - _extract_minus(term))
            else:
                unit_eqs.append(s.symbol - _extract_minus(expr))
        elif isinstance(s, ODESystem):
            amt_unit = di[di.typeix['dose'][0].name].unit
            time_unit = di[di.idv_column.name].unit
            for e in s.compartmental_matrix.diagonal():
                if e.is_Add:
                    for term in e.args:
                        unit_eqs.append(amt_unit / time_unit - _extract_minus(term))
                elif e == 0:
                    pass
                else:
                    unit_eqs.append(amt_unit / time_unit - _extract_minus(e))
            for a in s.amounts:
                unit_eqs.append(amt_unit - a)
    sol = sympy.solve(unit_eqs)
    return sol[symbol]



import numpy as np
import pandas as pd

import pharmpy.results
import pharmpy.tools.modelsearch.algorithms as algorithms
import pharmpy.tools.modelsearch.rankfuncs as rankfuncs
from pharmpy.modeling import summarize_modelfit_results
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow


def create_workflow(
    algorithm,
    mfl,
    rankfunc='ofv',
    cutoff=None,
    add_iivs=False,
    iiv_as_fullblock=False,
    add_mdt_iiv=False,
    model=None,
):
    algorithm_func = getattr(algorithms, algorithm)

    wf = Workflow()
    wf.name = 'modelsearch'

    if model:
        start_task = Task('start_modelsearch', start, model)
    else:
        start_task = Task('start_modelsearch', start)

    wf.add_task(start_task)

    if model and not model.modelfit_results:
        wf_fit = create_fit_workflow(n=1)
        wf.insert_workflow(wf_fit, predecessors=start_task)
        start_model_task = wf_fit.output_tasks
    else:
        start_model_task = [start_task]

    wf_search, candidate_model_tasks, model_features = algorithm_func(
        mfl, add_iivs, iiv_as_fullblock, add_mdt_iiv
    )
    wf.insert_workflow(wf_search, predecessors=wf.output_tasks)

    task_result = Task(
        'results',
        post_process_results,
        algorithm,
        rankfunc,
        cutoff,
        model_features,
    )

    wf.add_task(task_result, predecessors=start_model_task + candidate_model_tasks)

    return wf


def start(model):
    return model


def post_process_results(algorithm, rankfunc, cutoff, model_features, *models):
    res_models = []
    start_model = None
    for model in models:
        if not model.name.startswith('modelsearch_candidate'):
            start_model = model
        else:
            res_models.append(model)

    if not start_model:
        raise ValueError('Error in workflow: No starting model')

    summary_tool = create_summary(
        res_models, start_model, rankfunc, cutoff, model_features, algorithm=algorithm
    )

    best_model_name = summary_tool['rank'].idxmin()
    try:
        best_model = [model for model in res_models if model.name == best_model_name][0]
    except IndexError:
        best_model = start_model

    summary_models = summarize_modelfit_results([start_model] + res_models)

    res = ModelSearchResults(
        summary_tool=summary_tool,
        summary_models=summary_models,
        best_model=best_model,
        start_model=start_model,
        models=res_models,
    )

    return res


class ModelSearchResults(pharmpy.results.Results):
    def __init__(
        self, summary_tool=None, summary_models=None, best_model=None, start_model=None, models=None
    ):
        self.summary_tool = summary_tool
        self.summary_models = summary_models
        self.best_model = best_model
        self.start_model = start_model
        self.models = models


def create_summary(
    models,
    start_model,
    rankfunc_name,
    cutoff,
    model_features,
    rank_by_not_worse=False,
    bic_type='mixed',
    algorithm=None,
):
    res_data = {'parent_model': [], f'd{rankfunc_name}': [], 'features': [], 'rank': []}
    model_names = []
    models_all = [start_model] + models
    rankfunc = getattr(rankfuncs, rankfunc_name)
    kwargs = dict()
    if cutoff is not None:
        kwargs['cutoff'] = cutoff
    if rankfunc_name == 'ofv':
        kwargs['rank_by_not_worse'] = rank_by_not_worse
    if rankfunc_name == 'bic':
        kwargs['bic_type'] = bic_type
    ranks = rankfunc(start_model, models_all, **kwargs)
    delta_diff = rankfuncs.create_diff_dict(
        rankfunc_name, start_model, models_all, bic_type=bic_type
    )
    for model in models_all:
        model_names.append(model.name)
        res_data['parent_model'].append(model.parent_model)
        res_data[f'd{rankfunc_name}'].append(delta_diff[model.name])
        if model.name == start_model.name:
            res_data['features'].append(None)
        # FIXME: make more general
        elif algorithm == 'reduced_stepwise':
            feat = model_features[model.name]
            if model.parent_model in model_names and model.parent_model != start_model.name:
                idx = model_names.index(model.parent_model)
                feat_parent = res_data['features'][idx]
                feat_all = feat_parent + (feat,)
            else:
                feat_all = (feat,)
            res_data['features'].append(feat_all)
        else:
            res_data['features'].append(model_features[model.name])
        if model in ranks:
            res_data['rank'].append(ranks.index(model) + 1)
        else:
            res_data['rank'].append(np.nan)

    # FIXME: in ranks, if any row has NaN the rank converts to float
    return pd.DataFrame(res_data, index=model_names)



import shutil

import numpy as np
import pytest

from pharmpy import Model
from pharmpy.modeling import add_iiv, fit, run_tool
from pharmpy.utils import TemporaryDirectoryChanger


def test_exhaustive(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'
        res = run_tool(
            'modelsearch', 'exhaustive', 'ABSORPTION(ZO);PERIPHERALS(1)', model=model_start
        )

        assert len(res.summary_tool) == 4
        assert len(res.summary_models) == 4
        assert len(res.models) == 3
        assert all(
            model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
            for model in res.models
        )
        assert res.best_model.name == 'modelsearch_candidate1'
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 4
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()


@pytest.mark.parametrize(
    'mfl, no_of_models, best_model_name, last_model_parent_name',
    [
        ('ABSORPTION(ZO);PERIPHERALS(1)', 4, 'modelsearch_candidate2', 'modelsearch_candidate2'),
        ('ABSORPTION(ZO);ELIMINATION(ZO)', 4, 'mox2', 'modelsearch_candidate2'),
        ('ABSORPTION(ZO);TRANSITS(1)', 2, 'mox2', 'mox2'),
        (
            'ABSORPTION([ZO,SEQ-ZO-FO]);PERIPHERALS(1)',
            7,
            'modelsearch_candidate3',
            'modelsearch_candidate3',
        ),
        ('LAGTIME();TRANSITS(1)', 2, 'mox2', 'mox2'),
        ('ABSORPTION(ZO);TRANSITS(3, *)', 3, 'mox2', 'mox2'),
    ],
)
def test_exhaustive_stepwise_basic(
    tmp_path, testdata, mfl, no_of_models, best_model_name, last_model_parent_name
):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'
        res = run_tool('modelsearch', 'exhaustive_stepwise', mfl, model=model_start)

        assert len(res.summary_tool) == no_of_models + 1
        assert len(res.summary_models) == no_of_models + 1
        assert len(res.models) == no_of_models
        assert res.best_model.name == best_model_name

        assert res.models[0].parent_model == 'mox2'
        assert res.models[-1].parent_model == last_model_parent_name
        if last_model_parent_name != 'mox2':
            last_model_features = res.summary_tool.loc[res.models[-1].name]['features']
            parent_model_features = res.summary_tool.loc[last_model_parent_name]['features']
            assert last_model_features[: len(parent_model_features)] == parent_model_features

        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()


@pytest.mark.parametrize(
    'mfl, iiv_as_fullblock, add_mdt_iiv, no_of_models, best_model_name, no_of_added_etas',
    [
        ('ABSORPTION(ZO);PERIPHERALS(1)', False, False, 4, 'modelsearch_candidate2', 2),
        ('ABSORPTION(ZO);ELIMINATION(ZO)', False, False, 4, 'mox2', 1),
        ('ABSORPTION(ZO);ELIMINATION(MIX-FO-MM)', False, False, 4, 'modelsearch_candidate2', 2),
        ('ABSORPTION(ZO);PERIPHERALS([1, 2])', False, False, 8, 'modelsearch_candidate5', 4),
        ('LAGTIME();TRANSITS(1)', False, False, 2, 'mox2', 1),
        ('ABSORPTION(ZO);PERIPHERALS(1)', True, False, 4, 'modelsearch_candidate2', 2),
        ('PERIPHERALS(1);LAGTIME()', False, True, 4, 'modelsearch_candidate1', 1),
    ],
)
def test_exhaustive_stepwise_add_iivs(
    tmp_path,
    testdata,
    mfl,
    iiv_as_fullblock,
    add_mdt_iiv,
    no_of_models,
    best_model_name,
    no_of_added_etas,
):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'
        res = run_tool(
            'modelsearch',
            'exhaustive_stepwise',
            mfl,
            add_iivs=True,
            iiv_as_fullblock=iiv_as_fullblock,
            add_mdt_iiv=add_mdt_iiv,
            model=model_start,
        )

        import pandas as pd

        pd.set_option('display.max_columns', None)
        print(res.summary_tool)
        assert len(res.summary_tool) == no_of_models + 1
        assert len(res.summary_models) == no_of_models + 1
        assert len(res.models) == no_of_models
        assert res.best_model.name == best_model_name
        model_last = res.models[no_of_models - 1]
        assert (
            len(model_last.random_variables.etas) - len(model_start.random_variables.etas)
            == no_of_added_etas
        )
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()


def test_exhaustive_stepwise_already_fit(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'

        fit(model_start)

        mfl = 'ABSORPTION(ZO);PERIPHERALS(1)'
        res = run_tool('modelsearch', 'exhaustive_stepwise', mfl, model=model_start)

        assert len(res.summary_tool) == 5
        assert len(res.summary_models) == 5
        assert len(res.models) == 4
        assert all(
            model.modelfit_results and not np.isnan(model.modelfit_results.ofv)
            for model in res.models
        )
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 4
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()


def test_exhaustive_stepwise_start_model_fail(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'

        add_iiv(model_start, 'V', 'incorrect_syntax')

        mfl = 'ABSORPTION(ZO);PERIPHERALS(1)'
        res = run_tool('modelsearch', 'exhaustive_stepwise', mfl, model=model_start)

        assert len(res.summary_tool) == 5
        assert len(res.summary_models) == 5
        assert res.summary_tool['dofv'].isnull().values.all()
        assert len(res.models) == 4
        assert all(model.modelfit_results is None for model in res.models)
        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 5


@pytest.mark.parametrize(
    'mfl, no_of_models, best_model_name, last_model_parent_name',
    [
        (
            'ABSORPTION(ZO);LAGTIME();PERIPHERALS(1)',
            12,
            'modelsearch_candidate2',
            'modelsearch_candidate9',
        ),
    ],
)
def test_reduced_stepwise(
    tmp_path, testdata, mfl, no_of_models, best_model_name, last_model_parent_name
):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'
        res = run_tool('modelsearch', 'reduced_stepwise', mfl, model=model_start)

        assert len(res.summary_tool) == no_of_models + 1
        assert len(res.summary_models) == no_of_models + 1
        assert len(res.models) == no_of_models
        assert res.best_model.name == best_model_name

        assert res.models[0].parent_model == 'mox2'
        assert res.models[-1].parent_model == last_model_parent_name
        if last_model_parent_name != 'mox2':
            last_model_features = res.summary_tool.loc[res.models[-1].name]['features']
            parent_model_features = res.summary_tool.loc[last_model_parent_name]['features']
            assert last_model_features[: len(parent_model_features)] == parent_model_features

        rundir = tmp_path / 'modelsearch_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()



import pytest

from pharmpy import Model
from pharmpy.tools.estmethod.tool import _clear_estimation_steps, _create_est_model, create_workflow


@pytest.mark.parametrize(
    'methods, solvers, no_of_models',
    [
        ('foce', None, 2),
        (['foce', 'laplace'], None, 4),
        (['laplace'], None, 3),
        ('foce', ['lsoda'], 4),
        ('foce', 'all', 14),
    ],
)
def test_algorithm(methods, solvers, no_of_models):
    wf = create_workflow(methods=methods, solvers=solvers)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == no_of_models


@pytest.mark.parametrize(
    'method, est_rec, eval_rec',
    [
        (
            'fo',
            '$ESTIMATION METHOD=ZERO INTER MAXEVAL=9999 AUTO=1 PRINT=10',
            '$ESTIMATION METHOD=IMP INTER EONLY=1 MAXEVAL=9999 ISAMPLE=10000 NITER=10 PRINT=10',
        ),
        (
            'laplace',
            '$ESTIMATION METHOD=COND LAPLACE INTER MAXEVAL=9999 AUTO=1 PRINT=10',
            '$ESTIMATION METHOD=IMP LAPLACE INTER EONLY=1 MAXEVAL=9999 ISAMPLE=10000 '
            'NITER=10 PRINT=10',
        ),
    ],
)
def test_create_est_model(pheno_path, method, est_rec, eval_rec):
    model = Model.create_model(pheno_path)
    assert len(model.estimation_steps) == 1
    est_model = _create_est_model(method, None, update=False, model=model)
    assert len(est_model.estimation_steps) == 2
    assert est_model.name == f'estmethod_{method.upper()}_raw_inits'
    assert est_model.model_code.split('\n')[-5] == est_rec
    assert est_model.model_code.split('\n')[-4] == eval_rec


def test_clear_estimation_steps(pheno_path):
    model = Model.create_model(pheno_path)
    assert len(model.estimation_steps) == 1
    _clear_estimation_steps(model)
    assert len(model.estimation_steps) == 0



"""
===================
Generic Model class
===================

**Base class of all implementations.**

Inherit to *implement*, i.e. to define support for a specific model type.

Definitions
-----------
"""

import copy
import io
import pathlib
import warnings
from pathlib import Path

import pandas as pd
import sympy

from pharmpy.datainfo import ColumnInfo, DataInfo
from pharmpy.estimation import EstimationSteps
from pharmpy.parameter import Parameters
from pharmpy.plugins.utils import detect_model
from pharmpy.random_variables import RandomVariables
from pharmpy.statements import ModelStatements


class ModelError(Exception):
    """Exception for errors in model object"""

    pass


class ModelSyntaxError(ModelError):
    """Exception for Syntax errors in model code"""

    def __init__(self, msg='model syntax error'):
        super().__init__(msg)


class Model:
    """The Pharmpy model class"""

    def __init__(self):
        self.parameters = Parameters([])
        self.random_variables = RandomVariables([])
        self.statements = ModelStatements([])
        self.dependent_variable = sympy.Symbol('y')
        self.observation_transformation = self.dependent_variable
        self.modelfit_results = None
        self.parent_model = None

    def __repr__(self):
        return f'<Pharmpy model object {self.name}>'

    def _repr_html_(self):
        stat = self.statements._repr_html_()
        rvs = self.random_variables._repr_latex_()
        return f'<hr>{stat}<hr>${rvs}$<hr>{self.parameters._repr_html_()}<hr>'

    @property
    def name(self):
        """Name of the model"""
        return self._name

    @name.setter
    def name(self, value):
        if not isinstance(value, str):
            raise TypeError("Name of a model has to be of string type")
        self._name = value

    @property
    def filename_extension(self):
        """Filename extension of model file"""
        return self._filename_extension

    @filename_extension.setter
    def filename_extension(self, value):
        if not isinstance(value, str):
            raise TypeError("Filename extension has to be of string type")
        self._filename_extension = value

    @property
    def dependent_variable(self):
        """The model dependent variable, i.e. y"""
        return self._dependent_variable

    @dependent_variable.setter
    def dependent_variable(self, value):
        self._dependent_variable = value

    @property
    def observation_transformation(self):
        """Transformation to be applied to the observation data"""
        return self._observation_transformation

    @observation_transformation.setter
    def observation_transformation(self, value):
        self._observation_transformation = value

    @property
    def parameters(self):
        """Definitions of population parameters

        See :class:`pharmpy.Parameters`
        """
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        if isinstance(value, Parameters):
            inits = value.inits
        else:
            inits = value
        if isinstance(inits, pd.Series):
            inits = inits.to_dict()
        if inits and not self.random_variables.validate_parameters(inits):
            nearest = self.random_variables.nearest_valid_parameters(inits)
            if nearest != inits:
                before, after = self._compare_before_after_params(inits, nearest)
                warnings.warn(
                    f"Adjusting initial estimates to create positive semidefinite "
                    f"omega/sigma matrices.\nBefore adjusting:  {before}.\n"
                    f"After adjusting: {after}"
                )
                if isinstance(value, Parameters):
                    value.inits = nearest
                else:
                    value = nearest
            else:
                raise ValueError("New parameter inits are not valid")

        if not isinstance(value, Parameters):
            self._parameters.inits = value
        else:
            self._parameters = value

    @staticmethod
    def _compare_before_after_params(old, new):
        before = dict()
        after = dict()
        for key, value in old.items():
            if new[key] != value:
                before[key] = value
                after[key] = new[key]
        return before, after

    @property
    def random_variables(self):
        """Definitions of random variables

        See :class:`pharmpy.RandomVariables`
        """
        return self._random_variables

    @random_variables.setter
    def random_variables(self, value):
        if not isinstance(value, RandomVariables):
            raise TypeError("model.random_variables must be of RandomVariables type")
        self._random_variables = value

    @property
    def statements(self):
        """Definitions of model statements

        See :class:`pharmpy.ModelStatements`
        """
        return self._statements

    @statements.setter
    def statements(self, value):
        if not isinstance(value, ModelStatements):
            raise TypeError("model.statements must be of ModelStatements type")
        self._statements = value

    @property
    def estimation_steps(self):
        """Definitions of estimation steps

        See :class:`pharmpy.EstimationSteps`
        """
        return self._estimation_steps

    @estimation_steps.setter
    def estimation_steps(self, value):
        if not isinstance(value, EstimationSteps):
            raise TypeError("model.estimation_steps must be of EstimationSteps type")
        self._estimation_steps = value

    @property
    def datainfo(self):
        """Definitions of model statements

        See :class:`pharmpy.ModelStatements`
        """
        return self._datainfo

    @datainfo.setter
    def datainfo(self, value):
        if not isinstance(value, DataInfo):
            raise TypeError("model.datainfo must be of DataInfo type")
        self._datainfo = value

    @property
    def dataset(self):
        """Dataset connected to model"""
        return self._dataset

    @dataset.setter
    def dataset(self, value):
        self._dataset = value
        self.update_datainfo()

    @property
    def initial_individual_estimates(self):
        """Initial estimates for individual parameters"""
        return self._initial_individual_estimates

    @initial_individual_estimates.setter
    def initial_individual_estimates(self, value):
        self._initial_individual_estimates = value

    @property
    def modelfit_results(self):
        """Modelfit results for this model"""
        return self._modelfit_results

    @modelfit_results.setter
    def modelfit_results(self, value):
        self._modelfit_results = value

    @property
    def model_code(self):
        """Model type specific code"""
        raise NotImplementedError("Generic model does not implement the model_code property")

    @property
    def parent_model(self):
        """Name of parent model"""
        return self._parent_model

    @parent_model.setter
    def parent_model(self, value):
        self._parent_model = value

    def read_modelfit_results(self):
        """Read in modelfit results"""
        raise NotImplementedError("Read modelfit results not implemented for generic models")

    def update_datainfo(self):
        """Update model.datainfo for a new dataset"""
        colnames = self.dataset.columns
        try:
            curdi = self.datainfo
        except AttributeError:
            curdi = DataInfo()
        newdi = DataInfo()
        for colname in colnames:
            try:
                col = curdi[colname]
            except IndexError:
                col = ColumnInfo(colname)
            newdi.append(col)
        if curdi != newdi:
            # Remove path if dataset has been updated
            newdi.path = None
        self.datainfo = newdi

    def copy(self):
        """Create a deepcopy of the model object"""
        model_copy = copy.deepcopy(self)
        model_copy.parent_model = self.name
        return model_copy

    @staticmethod
    def create_model(obj=None, **kwargs):
        """Factory for creating a :class:`pharmpy.model` object from an object representing the model

        .. _path-like object: https://docs.python.org/3/glossary.html#term-path-like-object

        Parameters
        ----------
        obj
            `path-like object`_ pointing to the model file or an IO object.

        Returns
        -------
        Model
            Generic :class:`~pharmpy.generic.Model` if obj is None, otherwise appropriate
            implementation is invoked (e.g. NONMEM7 :class:`~pharmpy.plugins.nonmem.Model`).
        """
        if isinstance(obj, str):
            path = Path(obj)
        elif isinstance(obj, pathlib.Path):  # DO NOT change to Path! Will fail fakefs tests
            path = obj
        elif isinstance(obj, io.IOBase):
            path = None
        elif obj is None:
            return Model()
        else:
            raise ValueError("Unknown input type to Model constructor")
        if path is not None:
            with open(path, 'r', encoding='latin-1') as fp:
                code = fp.read()
        else:
            code = obj.read()
        model_class = detect_model(code)
        model = model_class(code, path, **kwargs)
        # Setup model database here
        # Read in model results here?
        # Set filename extension?
        return model



from pharmpy.tools.amd.funcs import create_start_model


def test_create_start_model(testdata):
    path = testdata / 'nonmem' / 'pheno.dta'
    model = create_start_model(path, modeltype='pk_iv')
    assert len(model.parameters) == 6
    assert 'POP_CL' in model.parameters
    assert 'POP_MAT' not in model.parameters

    model = create_start_model(path, modeltype='pk_oral')
    assert 'POP_CL' in model.parameters
    print(model.parameters)
    assert 'POP_MAT' in model.parameters



import shutil
from pathlib import Path

import pharmpy.modeling
from pharmpy import Model
from pharmpy.modeling import run_amd, run_iiv
from pharmpy.utils import TemporaryDirectoryChanger


def test_amd(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'pheno.dta', tmp_path)
        dipath = Path(pharmpy.modeling.__file__).parent / 'example_models' / 'pheno.datainfo'
        shutil.copy2(dipath, tmp_path)

        res = run_amd(tmp_path / 'pheno.dta', mfl='LAGTIME();PERIPHERALS(1)')
        assert res


def test_iiv(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'

        res = run_iiv(model_start)

        assert len(res.summary_tool) == 2
        assert len(res.summary_tool[0]) == 8
        assert len(res.summary_tool[1]) == 5
        assert len(res.summary_models) == 12
        assert len(res.models) == 11
        rundir1 = tmp_path / 'iiv_dir1'
        assert rundir1.is_dir()
        assert len(list((rundir1 / 'models').iterdir())) == 8
        rundir2 = tmp_path / 'iiv_dir2'
        assert rundir2.is_dir()
        assert len(list((rundir2 / 'models').iterdir())) == 4



from pharmpy.parameter import Parameter
from pharmpy.statements import Assignment, sympify

from .expressions import create_symbol


def add_allometry(
    model,
    allometric_variable='WT',
    reference_value=70,
    parameters=None,
    initials=None,
    lower_bounds=None,
    upper_bounds=None,
    fixed=True,
):
    """Add allometric scaling of parameters

    Add an allometric function to each listed parameter. The function will be
    P=P*(X/Z)**T where P is the parameter, X the allometric_variable, Z the reference_value
    and T is a theta. Default is to automatically use clearance and volume parameters.

    Parameters
    ----------
    model : Model
        Pharmpy model
    allometric_variable : str or Symbol
        Variable to use for allometry (X above)
    reference_value : str, int, float or expression
        Reference value (Z above)
    parameters : list
        Parameters to use or None (default) for all available CL, Q and V parameters
    initials : list
        Initial estimates for the exponents. Default is to use 0.75 for CL and Qs and 1 for Vs
    lower_bounds : list
        Lower bounds for the exponents. Default is 0 for all parameters
    upper_bounds : list
        Upper bounds for the exponents. Default is 2 for all parameters
    fixed : bool
        Whether the exponents should be fixed

    Return
    ------
    Model
        Pharmpy model object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, add_allometry
    >>> model = load_example_model("pheno")
    >>> add_allometry(model, allometric_variable='WGT')     # doctest: +ELLIPSIS
    <...>
    >>> model.statements.before_odes
             ⎧TIME  for AMT > 0
             ⎨
    BTIME := ⎩ 0     otherwise
    TAD := -BTIME + TIME
    TVCL := THETA(1)⋅WGT
    TVV := THETA(2)⋅WGT
           ⎧TVV⋅(THETA(3) + 1)  for APGR < 5
           ⎨
    TVV := ⎩       TVV           otherwise
                ETA(1)
    CL := TVCL⋅ℯ
                  ALLO_CL
             ⎛WGT⎞
          CL⋅⎜───⎟
    CL :=    ⎝ 70⎠
              ETA(2)
    V := TVV⋅ℯ
                ALLO_V
           ⎛WGT⎞
         V⋅⎜───⎟
    V :=   ⎝ 70⎠
    S₁ := V

    """
    allometric_variable = sympify(allometric_variable)
    reference_value = sympify(reference_value)

    odes = model.statements.ode_system
    central = odes.central_compartment
    output = odes.output_compartment

    if parameters is None:
        parameters = []
        inits = []
        elimination_rate = odes.get_flow(central, output)
        cl, vc = elimination_rate.as_numer_denom()
        if cl.is_Symbol and vc.is_Symbol:
            parameters += [cl, vc]
            inits += [0.75, 1.0]
        for periph in odes.peripheral_compartments:
            rate1 = odes.get_flow(central, periph)
            q1, v1 = rate1.as_numer_denom()
            if q1.is_Symbol and v1.is_Symbol:
                if q1 not in parameters:
                    parameters.append(q1)
                    inits.append(0.75)
                if v1 not in parameters:
                    parameters.append(v1)
                    inits.append(1.0)
            rate2 = odes.get_flow(periph, central)
            q2, v2 = rate2.as_numer_denom()
            if q2.is_Symbol and v2.is_Symbol:
                if q2 not in parameters:
                    parameters.append(q2)
                    inits.append(0.75)
                if v2 not in parameters:
                    parameters.append(v2)
                    inits.append(1.0)
        if initials is None:
            initials = inits
        if not parameters:
            raise ValueError("No parameters found")
    else:
        if not parameters:
            raise ValueError("No parameters provided")

        parameters = [sympify(p) for p in parameters]
        if initials is None:
            # Need to understand which parameter is CL or Q and which is V
            cls = []
            vcs = []
            elimination_rate = odes.get_flow(central, output)
            cl, vc = elimination_rate.as_numer_denom()
            if cl.is_Symbol and vc.is_Symbol:
                cls.append(cl)
                vcs.append(vc)
            for periph in odes.peripheral_compartments:
                rate1 = odes.get_flow(periph, central)
                q1, v1 = rate1.as_numer_denom()
                if q1.is_Symbol and v1.is_Symbol:
                    cls.append(q1)
                    vcs.append(v1)
                rate2 = odes.get_flow(periph, central)
                q2, v2 = rate2.as_numer_denom()
                if q2.is_Symbol and v2.is_Symbol:
                    cls.append(q2)
                    vcs.append(v2)

            initials = []
            for p in parameters:
                if p in cls:
                    initials.append(0.75)
                elif p in vcs:
                    initials.append(1.0)

    if lower_bounds is None:
        lower_bounds = [0] * len(parameters)
    if upper_bounds is None:
        upper_bounds = [2] * len(parameters)

    if not (len(parameters) == len(initials) == len(lower_bounds) == len(upper_bounds)):
        raise ValueError("The number of parameters, initials and bounds must be the same")

    for p, init, lower, upper in zip(parameters, initials, lower_bounds, upper_bounds):
        symb = create_symbol(model, f'ALLO_{p.name}')
        param = Parameter(symb.name, init=init, lower=lower, upper=upper, fix=fixed)
        model.parameters.append(param)
        expr = p * (allometric_variable / reference_value) ** param.symbol
        new_ass = Assignment(p, expr)
        p_ass = model.statements.find_assignment(p)
        model.statements.insert_after(p_ass, new_ass)

    return model



from functools import partial

import pharmpy.tools.scm as scm
from pharmpy.results import Results
from pharmpy.tools.amd.funcs import create_start_model
from pharmpy.workflows import default_tool_database

from .allometry import add_allometry
from .common import convert_model, read_model, write_model
from .data import remove_loq_data
from .run import fit, run_tool


class AMDResults(Results):
    def __init__(self, final_model=None):
        self.final_model = final_model


def run_amd(
    dataset_path,
    modeltype='pk_oral',
    cl_init=0.01,
    vc_init=1,
    mat_init=0.1,
    mfl=None,
    lloq=None,
    order=None,
    categorical=None,
    continuous=None,
):
    """Run Automatic Model Development (AMD) tool

    Runs structural modelsearch, IIV building, and resmod

    Parameters
    ----------
    dataset_path : Model
        Path to a dataset
    modeltype : str
        Type of model to build. Either 'pk_oral' or 'pk_iv'
    cl_init : float
        Initial estimate for the population clearance
    vc_init : float
        Initial estimate for the central compartment population volume
    mat_init : float
        Initial estimate for the mean absorption time (not for iv models)
    mfl : str
        MFL for search space for structural model
    lloq : float
        Lower limit of quantification. LOQ data will be removed.
    order : list
        Runorder of components
    categorical : list
        List of categorical covariates
    continuous : list
        List of continuouts covariates

    Returns
    -------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> run_amd(model)      # doctest: +SKIP

    See also
    --------
    run_iiv
    run_tool

    """

    model = create_start_model(
        dataset_path, modeltype=modeltype, cl_init=cl_init, vc_init=vc_init, mat_init=mat_init
    )
    model = convert_model(model, 'nonmem')  # FIXME: Workaround for results retrieval system

    if lloq is not None:
        remove_loq_data(model, lloq=lloq)

    default_order = ['structural', 'iiv', 'residual', 'allometry', 'covariates']
    if order is None:
        order = default_order

    if mfl is None:
        if modeltype == 'pk_oral':
            mfl = (
                'ABSORPTION([ZO,SEQ-ZO-FO]);'
                'ELIMINATION([MM,MIX-FO-MM]);'
                'LAGTIME();'
                'TRANSITS([1,3,10],*);'
                'PERIPHERALS([1,2])'
            )
        else:
            mfl = 'ELIMINATION([MM,MIX-FO-MM]);' 'PERIPHERALS([1,2])'

    db = default_tool_database(toolname='amd')
    # FIXME: Workaround for nonmem<->generic eta names gets lost somewhere
    write_model(model, db.path)
    model = read_model(db.path / (model.name + '.ctl'))
    fit(model)

    run_funcs = []
    for section in order:
        if section == 'structural':
            func = partial(_run_modelsearch, mfl=mfl, path=db.path)
            run_funcs.append(func)
        elif section == 'iiv':
            func = partial(_run_iiv, path=db.path)
            run_funcs.append(func)
        elif section == 'residual':
            func = partial(_run_resmod, path=db.path)
            run_funcs.append(func)
        elif section == 'allometry':
            func = partial(_run_allometry, allometric_variable=None)
            run_funcs.append(func)
        elif section == 'covariates':
            if scm.have_scm() and (continuous is not None or categorical is not None):
                func = partial(
                    _run_covariates, continuous=continuous, categorical=categorical, path=db.path
                )
                run_funcs.append(func)
        else:
            raise ValueError(
                f"Unrecognized section {section} in order. Must be one of {default_order}"
            )

    run_tool('modelfit', model, path=db.path / 'modelfit')

    next_model = model
    for func in run_funcs:
        next_model = func(next_model)

    res = AMDResults(final_model=next_model)
    return res


def _run_modelsearch(model, mfl, path):
    res_modelsearch = run_tool(
        'modelsearch', 'exhaustive_stepwise', mfl=mfl, model=model, path=path / 'modelsearch'
    )
    selected_model = res_modelsearch.best_model
    return selected_model


def _run_iiv(model, path):
    res_iiv = run_iiv(model, path=path)
    selected_iiv_model = res_iiv.best_model
    return selected_iiv_model


def _run_resmod(model, path):
    skip = []

    res_resmod = run_tool('resmod', model, path=path / 'resmod1')
    selected_model = res_resmod.best_model
    name = res_resmod.selected_model_name
    if name == 'base':
        return selected_model
    elif name[:12] == 'time_varying':
        skip.append('time_varying')
    else:
        skip.append(name)

    res_resmod = run_tool('resmod', selected_model, skip=skip, path=path / 'resmod2')
    selected_model = res_resmod.best_model
    name = res_resmod.selected_model_name
    if name == 'base':
        return selected_model
    elif name[:12] == 'time_varying':
        skip.append('time_varying')
    else:
        skip.append(name)

    res_resmod = run_tool('resmod', selected_model, skip=skip, path=path / 'resmod3')
    selected_model = res_resmod.best_model
    return selected_model


def _run_covariates(model, continuous, categorical, path):
    parameters = ['CL', 'VC', 'KMM', 'CLMM', 'MDT', 'MAT', 'QP1', 'QP2', 'VP1', 'VP2']

    if continuous is None:
        covariates = categorical
    elif categorical is None:
        covariates = continuous
    else:
        covariates = continuous + categorical

    relations = dict()
    for p in parameters:
        if model.statements.find_assignment(p):
            expr = model.statements.before_odes.full_expression(p)
            for eta in model.random_variables.etas:
                if eta.symbol in expr.free_symbols:
                    relations[p] = covariates
                    break
    res = scm.run_scm(model, relations, continuous=continuous, categorical=categorical, path=path)
    return res.final_model


def _run_allometry(model, allometric_variable):
    if allometric_variable is None:
        for col in model.datainfo:
            if col.descriptor == 'body weight':
                allometric_variable = col.name
                break

    if allometric_variable is not None:
        add_allometry(model, allometric_variable=allometric_variable)
        model.name = "scaled_model"
        fit(model)
    return model


def run_iiv(model, add_iivs=False, iiv_as_fullblock=False, rankfunc='ofv', cutoff=None, path=None):
    """Run IIV tool

    Runs two IIV workflows: testing the number of etas and testing which block structure

    Parameters
    ----------
    model : Model
        Pharmpy model
    add_iivs : bool
        Whether to add IIV on structural parameters. Default is False
    iiv_as_fullblock : bool
        Whether added etas should be as a fullblock. Default is False
    rankfunc : str
        Which ranking function should be used (OFV, AIC, BIC). Default is OFV
    cutoff : float
        Cutoff for which value of the ranking function that is considered significant. Default
        is 3.84
    path : Path
        Path of rundirectory

    Returns
    -------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> run_iiv(model)      # doctest: +SKIP

    See also
    --------
    run_amd
    run_tool

    """
    if path:
        path1 = path / 'iiv1'
        path2 = path / 'iiv2'
    else:
        path1 = path
        path2 = path

    res_no_of_etas = run_tool(
        'iiv',
        'brute_force_no_of_etas',
        add_iivs=add_iivs,
        iiv_as_fullblock=iiv_as_fullblock,
        rankfunc=rankfunc,
        cutoff=cutoff,
        model=model,
        path=path1,
    )
    res_block_structure = run_tool(
        'iiv',
        'brute_force_block_structure',
        rankfunc=rankfunc,
        cutoff=cutoff,
        model=res_no_of_etas.best_model,
        path=path2,
    )

    from pharmpy.modeling import summarize_modelfit_results

    summary_models = summarize_modelfit_results(
        [model] + res_no_of_etas.models + res_block_structure.models
    )

    from pharmpy.tools.iiv.tool import IIVResults

    res = IIVResults(
        summary_tool=[res_no_of_etas.summary_tool, res_block_structure.summary_tool],
        summary_models=summary_models,
        best_model=res_block_structure.best_model,
        models=res_no_of_etas.models + res_block_structure.models,
        start_model=model,
    )

    return res



import shutil
from os import stat
from pathlib import Path

from .baseclass import ModelDatabase


class LocalDirectoryDatabase(ModelDatabase):
    """ModelDatabase implementation for single local directory

    All files will be stored in the same directory. It is assumed that
    all files connected to a model are named modelname + extension. This means that
    care must be taken to keep filenames unique. Clashing filenames will
    be overwritten. It is recommended to use the LocalModelDirectoryDatabase
    instead.

    Parameters
    ----------
    path : str or Path
        Path to the database directory. Will be created if it does not exist.
    file_extension : str
        File extension to use for model files.
    """

    def __init__(self, path='.', file_extension='.mod'):
        path = Path(path)
        if not path.exists():
            path.mkdir(parents=True)
        self.path = path.resolve()
        self.file_extension = file_extension

    def store_local_file(self, model, path):
        if Path(path).is_file():
            shutil.copy2(path, self.path)

    def retrieve_local_files(self, name, destination_path):
        # Retrieve all files stored for one model
        files = self.path.glob(f'{name}.*')
        for f in files:
            shutil.copy2(f, destination_path)

    def retrieve_file(self, name, filename):
        # Return path to file
        path = self.path / filename
        if path.is_file() and stat(path).st_size > 0:
            return path
        else:
            raise FileNotFoundError(f"Cannot retrieve {filename} for {name}")

    def get_model(self, name):
        filename = name + self.file_extension
        path = self.path / filename
        from pharmpy.model import Model

        try:
            model = Model.create_model(path)
        except FileNotFoundError:
            raise KeyError('Model cannot be found in database')
        model.database = self
        model.read_modelfit_results()
        return model

    def __repr__(self):
        return f"LocalDirectoryDatabase({self.path})"


class LocalModelDirectoryDatabase(LocalDirectoryDatabase):
    """ModelDatabase implementation for a local directory structure

    Files will be stored in separate subdirectories named after each model.
    There are no restrictions on names of the files so models can have the same
    name of some connected file without creating a name clash.

    Parameters
    ----------
    path : str or Path
        Path to the base database directory. Will be created if it does not exist.
    file_extension : str
        File extension to use for model files.
    """

    def store_local_file(self, model, path):
        if Path(path).is_file():
            destination = self.path / model.name
            if not destination.is_dir():
                destination.mkdir(parents=True)
            shutil.copy2(path, destination)

    def retrieve_local_files(self, name, destination_path):
        path = self.path / name
        files = path.glob('*')
        for f in files:
            shutil.copy2(f, destination_path)

    def retrieve_file(self, name, filename):
        # Return path to file
        path = self.path / name / filename
        if path.is_file() and stat(path).st_size > 0:
            return path
        else:
            raise FileNotFoundError(f"Cannot retrieve {filename} for {name}")

    def get_model(self, name):
        filename = name + self.file_extension
        path = self.path / name / filename
        from pharmpy.model import Model

        model = Model.create_model(path)
        model.database = self
        model.read_modelfit_results()
        return model

    def __repr__(self):
        return f"LocalModelDirectoryDatabase({self.path})"



from functools import partial

import pandas as pd
import sympy
from scipy.stats import chi2

import pharmpy.model
import pharmpy.tools
from pharmpy import Parameter, Parameters, RandomVariable, RandomVariables
from pharmpy.estimation import EstimationStep, EstimationSteps
from pharmpy.modeling import (
    add_time_after_dose,
    create_symbol,
    get_mdv,
    set_combined_error_model,
    set_iiv_on_ruv,
    set_power_on_ruv,
)
from pharmpy.statements import Assignment, ModelStatements
from pharmpy.tools.modelfit import create_fit_workflow
from pharmpy.workflows import Task, Workflow

from ...modeling.error import remove_error_model, set_time_varying_error_model
from .results import calculate_results


def create_workflow(model=None, groups=4, p_value=0.05, skip=None):
    cutoff = float(chi2.isf(q=p_value, df=1))
    if skip is None:
        skip = []
    wf = Workflow()
    wf.name = "resmod"  # FIXME: Could have as input to Workflow

    if model is not None:
        start_task = Task('start_resmod', start, model)
    else:
        start_task = Task('start_resmod', start)

    task_base_model = Task('create_base_model', _create_base_model)
    wf.add_task(task_base_model, predecessors=start_task)

    tasks = []
    if 'IIV_on_RUV' not in skip:
        task_iiv = Task('create_iiv_on_ruv_model', _create_iiv_on_ruv_model)
        tasks.append(task_iiv)
        wf.add_task(task_iiv, predecessors=task_base_model)

    if 'power' not in skip and 'combined' not in skip:
        task_power = Task('create_power_model', _create_power_model)
        wf.add_task(task_power, predecessors=task_base_model)
        tasks.append(task_power)
        task_combined = Task('create_combined_error_model', _create_combined_model)
        wf.add_task(task_combined, predecessors=task_base_model)
        tasks.append(task_combined)

    if 'time_varying' not in skip:
        for i in range(1, groups):
            tvar = partial(_create_time_varying_model, groups=groups, i=i)
            task = Task(f"create_time_varying_model{i}", tvar)
            tasks.append(task)
            wf.add_task(task, predecessors=task_base_model)

    fit_wf = create_fit_workflow(n=1 + len(tasks))
    wf.insert_workflow(fit_wf, predecessors=[task_base_model] + tasks)
    post_pro = partial(post_process, cutoff=cutoff)
    task_post_process = Task('post_process', post_pro)
    wf.add_task(task_post_process, predecessors=[start_task] + fit_wf.output_tasks)

    task_unpack = Task('unpack', _unpack)
    wf.add_task(task_unpack, predecessors=[task_post_process])

    fit_final = create_fit_workflow(n=1)
    wf.insert_workflow(fit_final, predecessors=[task_unpack])
    _results = partial(_compare_full_models_results, cutoff=cutoff)
    task_results = Task('results', _results)
    wf.add_task(task_results, predecessors=[task_post_process] + fit_final.output_tasks)

    return wf


def start(model):
    return model


def post_process(start_model, *models, cutoff):
    res = calculate_results(
        base_model=_find_models(models)[0],
        tvar_models=_find_models(models)[1],
        other_models=_find_models(models)[2],
    )
    best_model = _create_best_model(start_model, res, cutoff=cutoff)
    res.best_model = best_model[0]
    res.selected_model_name = best_model[1]

    return res


def _unpack(res):
    return res.best_model


def _compare_full_models_results(res, start_model, cutoff):
    delta_ofv = start_model.modelfit_results.ofv - res.best_model.modelfit_results.ofv

    if delta_ofv <= cutoff:
        res.best_model = start_model
        res.selected_model_name = 'base'
    return res


def _find_models(models):
    base_model = None
    tvar_models = []
    other_models = []
    for model in models:
        if model.name == 'base':
            base_model = model
        elif model.name[:12] == 'time_varying':
            tvar_models.append(model)
        else:
            other_models.append(model)
    return base_model, tvar_models, other_models


def _create_base_model(input_model):
    base_model = pharmpy.model.Model()
    theta = Parameter('theta', 0.1)
    omega = Parameter('omega', 0.01, lower=0)
    sigma = Parameter('sigma', 1, lower=0)
    params = Parameters([theta, omega, sigma])
    base_model.parameters = params

    eta = RandomVariable.normal('eta', 'iiv', 0, omega.symbol)
    sigma = RandomVariable.normal('epsilon', 'ruv', 0, sigma.symbol)
    rvs = RandomVariables([eta, sigma])
    base_model.random_variables = rvs

    y = Assignment('Y', theta.symbol + eta.symbol + sigma.symbol)
    stats = ModelStatements([y])
    base_model.statements = stats

    base_model.dependent_variable = y.symbol
    base_model.name = 'base'
    base_model.dataset = _create_dataset(input_model)
    base_model.database = input_model.database

    est = EstimationStep('foce', interaction=True)
    base_model.estimation_steps = EstimationSteps([est])
    return base_model


def _create_iiv_on_ruv_model(input_model):
    base_model = input_model
    model = base_model.copy()
    set_iiv_on_ruv(model)
    model.name = 'IIV_on_RUV'
    return model


def _create_power_model(input_model):
    base_model = input_model
    model = base_model.copy()
    set_power_on_ruv(model, ipred='IPRED', lower_limit=None, zero_protection=True)
    model.name = 'power'
    return model


def _create_time_varying_model(input_model, groups, i):
    base_model = input_model
    model = base_model.copy()
    quantile = i / groups
    cutoff = model.dataset['TAD'].quantile(q=quantile)
    set_time_varying_error_model(model, cutoff=cutoff, idv='TAD')
    model.name = f"time_varying{i}"
    return model


def _create_combined_model(input_model):
    base_model = input_model
    model = base_model.copy()
    remove_error_model(model)
    s = model.statements[0]
    ruv_prop = create_symbol(model, 'epsilon_p')
    ruv_add = create_symbol(model, 'epsilon_a')
    ipred = sympy.Symbol('IPRED')
    s.expression = s.expression + ruv_prop + ruv_add / ipred

    sigma_prop = Parameter('sigma_prop', 1, lower=0)
    model.parameters.append(sigma_prop)
    model.dataset['IPRED'].replace(0, 2.225e-307, inplace=True)
    ipred_min = model.dataset['IPRED'].min()
    sigma_add_init = ipred_min / 2
    sigma_add = Parameter('sigma_add', sigma_add_init, lower=0)
    model.parameters.append(sigma_add)

    eps_prop = RandomVariable.normal(ruv_prop.name, 'ruv', 0, sigma_prop.symbol)
    model.random_variables.append(eps_prop)
    eps_add = RandomVariable.normal(ruv_add.name, 'ruv', 0, sigma_add.symbol)
    model.random_variables.append(eps_add)

    model.name = 'combined'
    return model


def _create_dataset(input_model):
    input_model = input_model.copy()
    residuals = input_model.modelfit_results.residuals
    cwres = residuals['CWRES'].reset_index(drop=True)
    predictions = input_model.modelfit_results.predictions
    if 'CIPREDI' in predictions:
        ipredcol = 'CIPREDI'
    elif 'IPRED' in predictions:
        ipredcol = 'IPRED'
    else:
        raise ValueError("Need CIPREDI or IPRED")
    ipred = predictions[ipredcol].reset_index(drop=True)
    mdv = get_mdv(input_model)
    mdv = mdv.reset_index(drop=True)
    label_id = input_model.datainfo.id_column.name
    input_id = input_model.dataset[label_id].astype('int64').squeeze().reset_index(drop=True)
    _time_after_dose(input_model)
    tad = input_model.dataset['TAD'].squeeze().reset_index(drop=True)
    df = pd.concat([mdv, input_id, tad, ipred], axis=1)
    df = df[df['MDV'] == 0].reset_index(drop=True)
    df = pd.concat([df, cwres], axis=1).rename(columns={'CWRES': 'DV', ipredcol: 'IPRED'})
    return df


def _time_after_dose(model):
    if 'TAD' in model.dataset:
        pass
    else:
        add_time_after_dose(model)
    return model


def _create_best_model(model, res, groups=4, cutoff=3.84):
    model = model.copy()
    selected_model_name = ''
    if any(res.models['dofv'] > cutoff):
        idx = res.models['dofv'].idxmax()
        name = idx[0]

        if name == 'power':
            set_power_on_ruv(model)
            model.parameters.inits = {
                'power1': res.models['parameters'].loc['power', 1, 1].get('theta') + 1
            }
        elif name == 'IIV_on_RUV':
            set_iiv_on_ruv(model)
            model.parameters.inits = {
                'IIV_RUV1': res.models['parameters'].loc['IIV_on_RUV', 1, 1].get('omega')
            }
        elif name[:12] == 'time_varying':
            _time_after_dose(model)
            i = int(name[-1])
            quantile = i / groups
            df = _create_dataset(model)
            tad = df['TAD']
            cutoff_tvar = tad.quantile(q=quantile)
            set_time_varying_error_model(model, cutoff=cutoff_tvar, idv='TAD')
            model.parameters.inits = {
                'time_varying': res.models['parameters']
                .loc[f"time_varying{i}", 1, 1]
                .get(f"theta_tvar{i}")
            }
        else:
            set_combined_error_model(model)
            model.parameters.inits = {
                'sigma_prop': res.models['parameters'].loc['combined', 1, 1].get('sigma_prop'),
                'sigma_add': res.models['parameters'].loc['combined', 1, 1].get('sigma_add'),
            }
        selected_model_name = name
        model.update_source()
    return model, selected_model_name



import copy
import importlib
import json
import lzma
from collections.abc import MutableSequence
from pathlib import Path

import altair as alt
import pandas as pd

import pharmpy.model


class ResultsJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pd.DataFrame) or isinstance(obj, pd.Series):
            if isinstance(obj, pd.DataFrame) and str(obj.columns.dtype) == 'int64':
                # Workaround for https://github.com/pandas-dev/pandas/issues/46392
                obj.columns = obj.columns.map(str)
            d = json.loads(obj.to_json(orient='table'))
            if isinstance(obj, pd.DataFrame):
                d['__class__'] = 'DataFrame'
            elif isinstance(obj, pd.Series):
                d['__class__'] = 'Series'
            else:
                d['__class__'] = obj.__class__.__name__
            return d
        elif obj.__class__.__module__.startswith('altair.'):
            d = obj.to_dict()
            d['__class__'] = 'vega-lite'
            return d
        elif isinstance(obj, pharmpy.model.Model):
            # TODO: consider using other representation, e.g. path
            return None
        else:
            return json.JSONEncoder.encode(self, obj)


class ResultsJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
        self.cls = None

    def object_hook(self, dct):
        if '__class__' in dct:
            cls = dct['__class__']
            del dct['__class__']
            if cls == 'DataFrame' or cls == 'Series':
                res = pd.read_json(json.dumps(dct), orient='table')
                return res
            elif cls == 'vega-lite':
                res = alt.Chart.from_dict(dct)
                return res
            else:
                self.cls = cls
        return dct


def read_results(path_or_buf):
    if '{' in str(path_or_buf):  # Heuristic to determine if path or buffer
        s = path_or_buf
    else:
        path = Path(path_or_buf)
        if path.is_dir():
            path /= 'results.json'
        if not path.is_file():
            raise FileNotFoundError(str(path))
        if path.name.endswith('.xz'):
            with lzma.open(path, 'r') as json_file:
                s = json_file.read().decode('utf-8')
        else:
            with open(path, 'r') as json_file:
                s = json_file.read()
    decoder = ResultsJSONDecoder()
    d = decoder.decode(s)
    tool_name = decoder.cls[:-7].lower()
    tool_module = importlib.import_module(f'pharmpy.tools.{tool_name}')
    results_class = tool_module.results_class
    res = results_class.from_dict(d)
    return res


class Results:
    """Base class for all result classes"""

    @classmethod
    def from_dict(cls, d):
        """Create results object from dictionary"""
        return cls(**d)

    def to_json(self, path=None, lzma=False):
        """Serialize results object as json

        Parameters
        ----------
        path : Path
            Path to save json file or None to serialize to string
        lzma : bool
            Set to compress file with lzma

        Returns
        -------
        str
            Json as string unless path was used
        """
        json_dict = self.to_dict()
        json_dict['__class__'] = self.__class__.__name__
        s = json.dumps(json_dict, cls=ResultsJSONEncoder)
        if path:
            if not lzma:
                with open(path, 'w') as fh:
                    fh.write(s)
            else:
                xz_path = path.parent / (path.name + '.xz')
                with lzma.open(xz_path, 'w') as fh:
                    fh.write(bytes(s, 'utf-8'))
        else:
            return s

    def get_and_reset_index(self, attr, **kwargs):
        """Wrapper to reset index of attribute or result from method.

        Used to facilitate importing multiindex dataframes into R
        """
        val = getattr(self, attr)
        if callable(val):
            df = val(**kwargs)
        else:
            df = val
        return df.reset_index()

    def to_dict(self):
        """Convert results object to a dictionary"""
        return vars(self).copy()

    def __str__(self):
        start = self.__class__.__name__
        s = f'{start}\n\n'
        d = self.to_dict()
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                s += value.to_string()
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    df = pd.DataFrame(value)
                    df_str = df.to_string(index=False)
                    df_str = df_str.split('\n')[1:]
                    s += '\n'.join(df_str)
            else:
                s += str(value) + '\n'
            s += '\n\n'
        return s

    def to_csv(self, path):
        """Save results as a human readable csv file

        Index will not be printed if it is a basic range.

        Parameters
        ----------
        path : Path
            Path to csv-file
        """
        d = self.to_dict()
        s = ""
        for key, value in d.items():
            if value.__class__.__module__.startswith('altair.'):
                continue
            elif isinstance(value, pharmpy.model.Model):
                continue
            elif isinstance(value, list) and isinstance(value[0], pharmpy.model.Model):
                continue
            s += f'{key}\n'
            if isinstance(value, pd.DataFrame):
                if isinstance(value.index, pd.RangeIndex):
                    use_index = False
                else:
                    use_index = True
                s += value.to_csv(index=use_index)
            elif isinstance(value, pd.Series):
                s += value.to_csv()
            elif isinstance(value, list):  # Print list of lists as table
                if len(value) > 0 and isinstance(value[0], list):
                    for row in value:
                        s += f'{",".join(map(str, row))}\n'
            else:
                s += str(value) + '\n'
            s += '\n'
        with open(path, 'w', newline='') as fh:
            print(s, file=fh)


class ModelfitResults(Results):
    """Base class for results from a modelfit operation

    model_name - name of model that generated the results model

    Attributes
    ----------
    correlation_matrix : pd.DataFrame
        Correlation matrix of the population parameter estimates
    covariance_matrix : pd.DataFrame
        Covariance matrix of the population parameter estimates
    information_matrix : pd.DataFrame
        Fischer information matrix of the population parameter estimates
    evaluation_ofv : float
        The objective function value as if the model was evaluated. Currently
        workfs for classical estimation methods by taking the OFV of the first
        iteration.
    individual_ofv : pd.Series
        OFV for each individual
    individual_estimates : pd.DataFrame
        Estimates for etas
    individual_estimates_covariance : pd.Series
        Estimated covariance between etas
    parameter_estimates : pd.Series
        Population parameter estimates
    parameter_estimates_sdcorr : pd.Series
        Population parameter estimates with variability parameters as standard deviations and
        correlations
    residuals: pd.DataFrame
        Table of various residuals
    estimation_runtime : float
        Runtime for one estimation step
    runtime_total : float
        Total runtime of estimation
    standard_errors : pd.Series
        Standard errors of the population parameter estimates
    standard_errors_sdcorr : pd.Series
        Standard errors of the population parameter estimates on standard deviation and correlation
        scale
    termination_cause : str
        The cause of premature termination. One of 'maxevals_exceeded' and 'rounding_errors'
    function_evaluations : int
        Number of function evaluations
    """

    def __init__(
        self,
        ofv=None,
        parameter_estimates=None,
        parameter_estimates_sdcorr=None,
        covariance_matrix=None,
        correlation_matrix=None,
        information_matrix=None,
        standard_errors=None,
        minimization_successful=None,
        individual_ofv=None,
        individual_estimates=None,
        individual_estimates_covariance=None,
        residuals=None,
        runtime_total=None,
        termination_cause=None,
        function_evaluations=None,
    ):
        self.ofv = ofv
        self.parameter_estimates = parameter_estimates
        self.parameter_estimates_sdcorr = parameter_estimates_sdcorr
        self.covariance_matrix = covariance_matrix
        self.correlation_matrix = correlation_matrix
        self.standard_errors = standard_errors
        self.minimization_successful = minimization_successful
        self.individual_estimates = individual_estimates
        self.individual_ofv = individual_ofv
        self.residuals = residuals
        self.runtime_total = runtime_total
        self.termination_cause = termination_cause
        self.function_evaluations = function_evaluations

    def __bool__(self):
        return bool(self.ofv) and bool(self.parameter_estimates)

    def to_dict(self):
        """Convert results object to a dictionary"""
        return {'parameter_estimates': self.parameter_estimates}

    @property
    def relative_standard_errors(self):
        """Relative standard errors of population parameter estimates"""
        if self.standard_errors is not None:
            ser = self.standard_errors / self.parameter_estimates
            ser.name = 'RSE'
            return ser


class ChainedModelfitResults(MutableSequence, ModelfitResults):
    """A sequence of modelfit results given in order from first to final
    inherits from both list and ModelfitResults. Each method from ModelfitResults
    will be performed on the final modelfit object
    """

    def __init__(self, results=None):
        if isinstance(results, ChainedModelfitResults):
            self._results = copy.deepcopy(results._results)
        elif results is None:
            self._results = []
        else:
            self._results = list(results)

    def __getitem__(self, ind):
        return self._results[ind]

    def __setitem__(self, ind, value):
        self._results[ind] = value

    def __delitem__(self, ind):
        del self._results[ind]

    def __len__(self):
        return len(self._results)

    def insert(self, ind, value):
        self._results.insert(ind, value)

    @property
    def ofv(self):
        return self[-1].ofv

    @property
    def evaluation_ofv(self):
        return self[0].evaluation_ofv

    @property
    def minimization_successful(self):
        return self._get_last_est('minimization_successful')

    @property
    def estimation_runtime(self):
        return self._get_last_est('estimation_runtime')

    def _get_last_est(self, attr):
        est_steps = self.model.estimation_steps
        # Find last estimation
        for i in reversed(range(len(self))):
            if not est_steps[i].evaluation and getattr(self[i], attr) is not None:
                return getattr(self[i], attr)
        # If all steps were evaluation the last evaluation step is relevant
        return getattr(self[-1], attr)

    @property
    def parameter_estimates(self):
        return self[-1].parameter_estimates

    @parameter_estimates.setter
    def parameter_estimates(self, value):
        self[-1].parameter_estimates = value

    @property
    def parameter_estimates_sdcorr(self):
        return self[-1].parameter_estimates_sdcorr

    @property
    def covariance_matrix(self):
        return self[-1].covariance_matrix

    @property
    def information_matrix(self):
        return self[-1].information_matrix

    @property
    def correlation_matrix(self):
        return self[-1].correlation_matrix

    @property
    def standard_errors(self):
        return self[-1].standard_errors

    @property
    def standard_errors_sdcorr(self):
        return self[-1].standard_errors_sdcorr

    @property
    def individual_ofv(self):
        return self[-1].individual_ofv

    @property
    def individual_estimates(self):
        return self[-1].individual_estimates

    @property
    def individual_estimates_covariance(self):
        return self[-1].individual_estimates_covariance

    @property
    def residuals(self):
        return self[-1].residuals

    @property
    def predictions(self):
        return self[-1].predictions

    @property
    def model_name(self):
        return self[-1].model_name

    @property
    def function_evaluations(self):
        return self._get_last_est('function_evaluations')

    @property
    def termination_cause(self):
        return self._get_last_est('termination_cause')

    @property
    def runtime_total(self):
        return self[-1].runtime_total

    def __repr__(self):
        return repr(self._results[-1])



from io import StringIO

import numpy as np
import pytest

from pharmpy import Model
from pharmpy.modeling import add_iiv, add_peripheral_compartment, create_joint_distribution
from pharmpy.tools.iiv.algorithms import (
    _get_possible_iiv_blocks,
    _is_current_block_structure,
    brute_force_block_structure,
)
from pharmpy.tools.iiv.tool import _update_inits_start_model


@pytest.mark.parametrize(
    'list_of_parameters, block_structure, no_of_models',
    [([], [], 4), (['QP1'], [], 14), ([], ['ETA(1)', 'ETA(2)'], 4)],
)
def test_brute_force_block_structure(testdata, list_of_parameters, block_structure, no_of_models):
    model = Model.create_model(testdata / 'nonmem' / 'models' / 'mox2.mod')
    add_peripheral_compartment(model)
    add_iiv(model, list_of_parameters, 'add')
    if block_structure:
        create_joint_distribution(model, block_structure)

    iivs = model.random_variables.iiv
    wf, model_features = brute_force_block_structure(iivs)
    fit_tasks = [task.name for task in wf.tasks if task.name.startswith('run')]

    assert len(fit_tasks) == no_of_models
    assert block_structure not in model_features.items()


def test_get_iiv_combinations_4_etas(testdata, pheno_path):
    model = Model.create_model(
        StringIO(
            '''
$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V*EXP(ETA(3))*EXP(ETA(4))

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA DIAGONAL(4)
 0.0309626  ;       IVCL
 0.031128  ;        IVV
 0.031128
 0.031128
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION
'''
        )
    )
    model.path = testdata / 'nonmem' / 'pheno.mod'  # To be able to find dataset

    iivs = model.random_variables.iiv
    iiv_single_block, iiv_multi_block = _get_possible_iiv_blocks(iivs)

    assert iiv_single_block == [
        ['ETA(1)', 'ETA(2)'],
        ['ETA(1)', 'ETA(3)'],
        ['ETA(1)', 'ETA(4)'],
        ['ETA(2)', 'ETA(3)'],
        ['ETA(2)', 'ETA(4)'],
        ['ETA(3)', 'ETA(4)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)'],
        ['ETA(1)', 'ETA(2)', 'ETA(4)'],
        ['ETA(1)', 'ETA(3)', 'ETA(4)'],
        ['ETA(2)', 'ETA(3)', 'ETA(4)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)', 'ETA(4)'],
    ]

    assert iiv_multi_block == [
        [['ETA(1)', 'ETA(2)'], ['ETA(3)', 'ETA(4)']],
        [['ETA(1)', 'ETA(3)'], ['ETA(2)', 'ETA(4)']],
        [['ETA(1)', 'ETA(4)'], ['ETA(2)', 'ETA(3)']],
    ]


def test_get_iiv_combinations_5_etas(testdata, pheno_path):
    model = Model.create_model(
        StringIO(
            '''
$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V*EXP(ETA(3))*EXP(ETA(4))*EXP(ETA(5))

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA DIAGONAL(5)
 0.0309626  ;       IVCL
 0.031128  ;        IVV
 0.031128
 0.031128
 0.031128
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION
'''
        )
    )

    model.path = testdata / 'nonmem' / 'pheno.mod'  # To be able to find dataset

    iivs = model.random_variables.iiv
    iiv_single_block, iiv_multi_block = _get_possible_iiv_blocks(iivs)

    assert iiv_single_block == [
        ['ETA(1)', 'ETA(2)'],
        ['ETA(1)', 'ETA(3)'],
        ['ETA(1)', 'ETA(4)'],
        ['ETA(1)', 'ETA(5)'],
        ['ETA(2)', 'ETA(3)'],
        ['ETA(2)', 'ETA(4)'],
        ['ETA(2)', 'ETA(5)'],
        ['ETA(3)', 'ETA(4)'],
        ['ETA(3)', 'ETA(5)'],
        ['ETA(4)', 'ETA(5)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)'],
        ['ETA(1)', 'ETA(2)', 'ETA(4)'],
        ['ETA(1)', 'ETA(2)', 'ETA(5)'],
        ['ETA(1)', 'ETA(3)', 'ETA(4)'],
        ['ETA(1)', 'ETA(3)', 'ETA(5)'],
        ['ETA(1)', 'ETA(4)', 'ETA(5)'],
        ['ETA(2)', 'ETA(3)', 'ETA(4)'],
        ['ETA(2)', 'ETA(3)', 'ETA(5)'],
        ['ETA(2)', 'ETA(4)', 'ETA(5)'],
        ['ETA(3)', 'ETA(4)', 'ETA(5)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)', 'ETA(4)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)', 'ETA(5)'],
        ['ETA(1)', 'ETA(2)', 'ETA(4)', 'ETA(5)'],
        ['ETA(1)', 'ETA(3)', 'ETA(4)', 'ETA(5)'],
        ['ETA(2)', 'ETA(3)', 'ETA(4)', 'ETA(5)'],
        ['ETA(1)', 'ETA(2)', 'ETA(3)', 'ETA(4)', 'ETA(5)'],
    ]

    assert iiv_multi_block == [
        [['ETA(1)', 'ETA(2)'], ['ETA(3)', 'ETA(4)', 'ETA(5)']],
        [['ETA(1)', 'ETA(3)'], ['ETA(2)', 'ETA(4)', 'ETA(5)']],
        [['ETA(1)', 'ETA(4)'], ['ETA(2)', 'ETA(3)', 'ETA(5)']],
        [['ETA(1)', 'ETA(5)'], ['ETA(2)', 'ETA(3)', 'ETA(4)']],
        [['ETA(2)', 'ETA(3)'], ['ETA(1)', 'ETA(4)', 'ETA(5)']],
        [['ETA(2)', 'ETA(4)'], ['ETA(1)', 'ETA(3)', 'ETA(5)']],
        [['ETA(2)', 'ETA(5)'], ['ETA(1)', 'ETA(3)', 'ETA(4)']],
        [['ETA(3)', 'ETA(4)'], ['ETA(1)', 'ETA(2)', 'ETA(5)']],
        [['ETA(3)', 'ETA(5)'], ['ETA(1)', 'ETA(2)', 'ETA(4)']],
        [['ETA(4)', 'ETA(5)'], ['ETA(1)', 'ETA(2)', 'ETA(3)']],
    ]


def test_is_current_block_structure(testdata):
    model_code = StringIO(
        '''
$PROBLEM PHENOBARB SIMPLE MODEL
$DATA pheno.dta IGNORE=@
$INPUT ID TIME AMT WGT APGR DV
$SUBROUTINE ADVAN1 TRANS2

$PK
CL=THETA(1)*EXP(ETA(1))
V=THETA(2)*EXP(ETA(2))
S1=V*EXP(ETA(3))*EXP(ETA(4))

$ERROR
Y=F+F*EPS(1)

$THETA (0,0.00469307) ; TVCL
$THETA (0,1.00916) ; TVV
$OMEGA DIAGONAL(4)
 0.0309626  ;       IVCL
 0.031128  ;        IVV
 0.031128
 0.031128
$SIGMA 0.013241

$ESTIMATION METHOD=1 INTERACTION
'''
    )
    model = Model.create_model(model_code)
    model.path = testdata / 'nonmem' / 'pheno.mod'  # To be able to find dataset
    list_of_etas_12 = ['ETA(1)', 'ETA(2)']
    create_joint_distribution(model, list_of_etas_12)
    assert _is_current_block_structure(model.random_variables.iiv, list_of_etas_12)
    list_of_etas_34 = ['ETA(3)', 'ETA(4)']
    assert not _is_current_block_structure(model.random_variables.iiv, list_of_etas_34)
    list_of_etas_23 = ['ETA(2)', 'ETA(3)']
    assert not _is_current_block_structure(model.random_variables.iiv, list_of_etas_23)
    create_joint_distribution(model)
    assert _is_current_block_structure(
        model.random_variables.iiv, list_of_etas_12 + list_of_etas_34
    )


def test_update_inits_start_model(pheno_path):
    model = Model.create_model(pheno_path)
    model.modelfit_results.parameter_estimates['OMEGA(1,1)'] = np.nan
    _update_inits_start_model(model)



import shutil

import pytest

from pharmpy import Model
from pharmpy.modeling import (
    add_iiv,
    add_peripheral_compartment,
    create_joint_distribution,
    run_tool,
    set_seq_zo_fo_absorption,
)
from pharmpy.utils import TemporaryDirectoryChanger


@pytest.mark.parametrize(
    'list_of_parameters, no_of_models, best_model_name',
    [([], 4, 'iiv_block_structure_candidate3'), (['QP1'], 14, 'iiv_block_structure_candidate10')],
)
def test_iiv_block_structure(tmp_path, testdata, list_of_parameters, no_of_models, best_model_name):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'

        add_peripheral_compartment(model_start)
        add_iiv(model_start, list_of_parameters, 'add')

        res = run_tool('iiv', 'brute_force_block_structure', model=model_start)

        assert len(res.summary_tool) == no_of_models + 1
        assert len(res.summary_models) == no_of_models + 1
        assert len(res.models) == no_of_models
        assert all(int(model.modelfit_results.ofv) in range(-2200, -2000) for model in res.models)
        assert all(model.random_variables != model_start.random_variables for model in res.models)
        assert res.best_model.name == best_model_name
        rundir = tmp_path / 'iiv_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == no_of_models + 1


def test_iiv_no_of_etas(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'

        res = run_tool('iiv', 'brute_force_no_of_etas', model=model_start)

        assert len(res.summary_tool) == 8
        assert len(res.summary_models) == 8
        assert len(res.models) == 7
        assert res.best_model.name == 'mox2'
        rundir = tmp_path / 'iiv_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 8


@pytest.mark.filterwarnings("ignore::UserWarning")
@pytest.mark.parametrize(
    'iiv_as_fullblock, best_model_name',
    [(False, 'iiv_no_of_etas_candidate4'), (True, 'iiv_no_of_etas_candidate4')],
)
def test_iiv_no_of_etas_added_iiv(tmp_path, testdata, iiv_as_fullblock, best_model_name):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'

        set_seq_zo_fo_absorption(model_start)
        res = run_tool(
            'iiv',
            'brute_force_no_of_etas',
            add_iivs=True,
            iiv_as_fullblock=iiv_as_fullblock,
            rankfunc='bic',
            model=model_start,
        )

        assert (len(res.start_model.random_variables['ETA(1)'].joint_names) > 0) is iiv_as_fullblock
        assert len(res.summary_tool) == 16
        assert len(res.summary_models) == 16
        assert len(res.models) == 15
        assert res.best_model.name == best_model_name
        rundir = tmp_path / 'iiv_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 16


def test_iiv_no_of_etas_fullblock(tmp_path, testdata):
    with TemporaryDirectoryChanger(tmp_path):
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mox2.mod', tmp_path)
        shutil.copy2(testdata / 'nonmem' / 'models' / 'mx19B.csv', tmp_path)
        model_start = Model.create_model('mox2.mod')
        model_start.datainfo.path = tmp_path / 'mx19B.csv'

        create_joint_distribution(model_start)
        res = run_tool('iiv', 'brute_force_no_of_etas', model=model_start)

        assert len(res.summary_tool) == 8
        assert len(res.summary_models) == 8
        assert len(res.models) == 7
        assert res.best_model.name == 'mox2'
        rundir = tmp_path / 'iiv_dir1'
        assert rundir.is_dir()
        assert len(list((rundir / 'models').iterdir())) == 8



from pharmpy.modeling import (
    add_time_after_dose,
    check_dataset,
    drop_columns,
    drop_dropped_columns,
    expand_additional_doses,
    get_concentration_parameters_from_data,
    get_covariate_baselines,
    get_doseid,
    get_doses,
    get_ids,
    get_mdv,
    get_number_of_individuals,
    get_number_of_observations,
    get_number_of_observations_per_individual,
    get_observations,
    list_time_varying_covariates,
    load_example_model,
    read_model,
    remove_loq_data,
    translate_nmtran_time,
    undrop_columns,
)

model = load_example_model("pheno")


def test_get_ids():
    assert get_ids(model) == list(range(1, 60))


def test_get_doseid():
    doseid = get_doseid(model)
    assert len(doseid) == 744
    assert doseid[0] == 1
    assert doseid[743] == 13

    # Same timepoint for dose and observation
    newmod = model.copy()
    newmod.dataset.loc[742, 'TIME'] = newmod.dataset.loc[743, 'TIME']
    doseid = get_doseid(newmod)
    assert len(doseid) == 744
    assert doseid[743] == 12
    assert doseid[742] == 13


def test_get_number_of_individuals():
    assert get_number_of_individuals(model) == 59


def test_get_observations():
    ser = get_observations(model)
    assert ser.loc[1, 2.0] == 17.3
    assert ser.loc[2, 63.5] == 24.6
    assert len(ser) == 155


def test_number_of_observations():
    assert get_number_of_observations(model) == 155
    assert list(get_number_of_observations_per_individual(model)) == [
        2,
        3,
        3,
        3,
        3,
        3,
        3,
        3,
        4,
        3,
        1,
        3,
        2,
        4,
        2,
        3,
        3,
        4,
        3,
        3,
        3,
        2,
        3,
        3,
        6,
        2,
        2,
        1,
        1,
        2,
        1,
        3,
        2,
        2,
        2,
        3,
        2,
        4,
        3,
        2,
        3,
        2,
        1,
        3,
        3,
        1,
        1,
        5,
        3,
        4,
        3,
        3,
        2,
        4,
        1,
        1,
        2,
        3,
        3,
    ]


def test_covariate_baselines():
    model.datainfo[['WGT', 'APGR']].types = 'covariate'
    df = get_covariate_baselines(model)
    assert len(df) == 59
    assert list(df.columns) == ['WGT', 'APGR']
    assert df.index.name == 'ID'
    assert df['WGT'].loc[2] == 1.5
    assert df['APGR'].loc[11] == 7.0


def test_doses():
    ser = get_doses(model)
    assert len(ser) == 589
    assert ser.loc[1, 0.0] == 25.0


def test_timevarying_covariates():
    a = list_time_varying_covariates(model)
    assert a == []


def test_get_mdv():
    mdv = get_mdv(model)
    label_test = model.datainfo.typeix['dose'][0].name
    data_test = model.dataset[label_test].astype('float64').squeeze()
    mdv_test = data_test.where(data_test == 0, other=1).astype('int64')
    result = mdv.equals(other=mdv_test)
    assert result is True


def test_add_time_after_dose():
    m = model.copy()
    add_time_after_dose(m)
    tad = m.dataset['TAD']
    assert tad[0] == 0.0
    assert tad[1] == 2.0
    assert tad[743] == 2.0


def test_get_concentration_parameters_from_data():
    df = get_concentration_parameters_from_data(model)
    assert df['Cmax'].loc[1, 1] == 17.3


def test_drop_columns():
    m = model.copy()
    drop_columns(m, "APGR")
    correct = ['ID', 'TIME', 'AMT', 'WGT', 'DV']
    assert m.datainfo.names == correct
    assert list(m.dataset.columns) == correct + ['FA1', 'FA2']
    drop_columns(m, ['DV', 'ID'])
    assert m.datainfo.names == ['TIME', 'AMT', 'WGT']
    assert list(m.dataset.columns) == ['TIME', 'AMT', 'WGT', 'FA1', 'FA2']
    drop_columns(m, ['TIME'], mark=True)
    assert m.datainfo['TIME'].drop
    assert list(m.dataset.columns) == ['TIME', 'AMT', 'WGT', 'FA1', 'FA2']


def test_drop_dropped_columns():
    m = model.copy()
    drop_dropped_columns(m)
    correct = ['ID', 'TIME', 'AMT', 'WGT', 'APGR', 'DV']
    assert list(m.dataset.columns) == correct
    drop_columns(m, ['ID', 'TIME', 'AMT'], mark=True)
    drop_dropped_columns(m)
    assert list(m.dataset.columns) == ['WGT', 'APGR', 'DV']


def test_undrop_columns():
    m = model.copy()
    drop_columns(m, ["APGR", "WGT"], mark=True)
    undrop_columns(m, "WGT")
    assert not m.datainfo["WGT"].drop
    assert m.datainfo["APGR"].drop


def test_remove_loq_data():
    m = model.copy()
    remove_loq_data(m, lloq=10, uloq=40)
    assert len(m.dataset) == 736


def test_check_dataset():
    m = model.copy()
    check_dataset(m)

    df = check_dataset(m, verbose=True, dataframe=True)
    assert df[df['code'] == 'A1']['result'].iloc[0] == 'OK'
    assert df[df['code'] == 'A4']['result'].iloc[0] == 'SKIP'

    m.dataset.loc[743, 'WGT'] = -1
    df = check_dataset(m, verbose=True, dataframe=True)
    assert df[df['code'] == 'A3']['result'].iloc[0] == 'FAIL'


def test_nmtran_time():
    m = load_example_model("pheno_linear")
    translate_nmtran_time(m)


def test_expand_additional_doses(testdata):
    model = read_model(testdata / 'nonmem' / 'models' / 'pef.mod')
    expand_additional_doses(model)
    df = model.dataset
    assert len(df) == 1494
    assert len(df.columns) == 5
    assert df.loc[0, 'AMT'] == 400.0
    assert df.loc[1, 'AMT'] == 400.0
    assert df.loc[2, 'AMT'] == 400.0
    assert df.loc[3, 'AMT'] == 400.0
    assert df.loc[4, 'AMT'] == 200.0
    assert df.loc[0, 'TIME'] == 0.0
    assert df.loc[1, 'TIME'] == 12.0
    assert df.loc[2, 'TIME'] == 24.0
    assert df.loc[3, 'TIME'] == 36.0
    assert df.loc[4, 'TIME'] == 48.0

    model = read_model(testdata / 'nonmem' / 'models' / 'pef.mod')
    expand_additional_doses(model, flag=True)
    df = model.dataset
    assert len(df) == 1494
    assert len(df.columns) == 6
    print(df)
    assert df.loc[0, 'EXPANDED']
    assert df.loc[1, 'EXPANDED']
    assert df.loc[2, 'EXPANDED']
    assert df.loc[3, 'EXPANDED']
    assert not df.loc[4, 'EXPANDED']



def create_report(results, path):
    """Create standard report for results

    The report will be an html created at specified path.

    Parameters
    ----------
    results : Results
        Results for which to create report
    path : Path
        Path to report file
    """
    import pharmpy.reporting.reporting as reporting

    reporting.generate_report(results.rst_path, path)



import copy
import uuid

import networkx as nx


class Workflow:
    """Workflow class

    Representation of a directed acyclic graph with Tasks as nodes and
    the flow of parameters as edges.

    Parameters
    ----------
    tasks : list
        List of tasks for initialization
    """

    def __init__(self, tasks=None):
        self._g = nx.DiGraph()
        if tasks:
            for task in tasks:
                self.add_task(task)

    def add_task(self, task, predecessors=None):
        """Add a task to the workflow

        Predecessors will be connected if given.

        Parameters
        ----------
        task : Task
            Task to add
        predecessors : list or Task
            One or multiple predecessor tasks to connect to the added task
        """
        self._g.add_node(task)
        if predecessors is not None:
            if not isinstance(predecessors, list):
                self._g.add_edge(predecessors, task)
            else:
                for pred in predecessors:
                    self._g.add_edge(pred, task)

    def insert_workflow(self, other, predecessors=None):
        """Insert other workflow

        Parameters
        ----------
        other : Workflow
            Workflow to insert
        predecessors : list or Task
            One or multiple predecessor tasks to connect to the inputs
            of the inserted workflow. If None all output tasks will be found
            and used as predecessors.
        """
        if predecessors is None:
            output_tasks = self.output_tasks
        else:
            if isinstance(predecessors, list):
                output_tasks = predecessors
            else:
                output_tasks = [predecessors]
        input_tasks = other.input_tasks
        self._g = nx.compose(self._g, other._g)
        if len(input_tasks) == len(output_tasks):
            for inp, outp in zip(input_tasks, output_tasks):
                self._g.add_edge(outp, inp)
        elif len(input_tasks) == 1:
            for outp in output_tasks:
                self._g.add_edge(outp, input_tasks[0])
        elif len(output_tasks) == 1:
            for inp in input_tasks:
                self._g.add_edge(output_tasks[0], inp)
        else:
            raise ValueError('Having N:M connections between workflows is currently not supported')

    def as_dask_dict(self):
        """Create a dask workflow dictionary

        Returns
        -------
        dict
            Dask workflow dictionary
        """
        ids = dict()
        for task in self._g.nodes():
            ids[task] = f'{task.name}-{uuid.uuid4()}'

        if len(self.output_tasks) == 1:
            ids[self.output_tasks[0]] = 'results'
        else:
            raise ValueError("Workflow can only have one output task")

        as_dict = dict()
        for task in nx.dfs_tree(self._g):
            key = ids[task]
            if task.has_input():
                input_list = list(task.task_input)
            else:
                input_list = []
            input_list.extend([ids[t] for t in self._g.predecessors(task)])
            value = (task.function, *input_list)
            as_dict[key] = value
        return as_dict

    @property
    def input_tasks(self):
        """All input (source) tasks of entire workflow"""
        return [node for node in self._g.nodes if self._g.in_degree(node) == 0]

    @property
    def output_tasks(self):
        """All output tasks (sink) tasks of entire workflow"""
        return [node for node in self._g.nodes if self._g.out_degree(node) == 0]

    def get_upstream_tasks(self, task):
        """Get all tasks upstream of a certain task

        Parameters
        ----------
        task : Task
            Downstream task

        Returns
        -------
        list
            Upstream tasks
        """
        edges = list(nx.edge_dfs(self._g, task, orientation='reverse'))
        return [node for node, _, _ in edges]

    def get_predecessors(self, task):
        """Get all predecessors of task

        Parameters
        ----------
        task : Task
            Downstream task

        Returns
        -------
        list
            Predecessors of task
        """
        return list(self._g.predecessors(task))

    def get_successors(self, task):
        """Get all successors of task

        Parameters
        ----------
        task : Task
            Upstream task

        Returns
        -------
        list
            Successors of task
        """
        return list(self._g.successors(task))

    def copy(self):
        """Deepcopy of workflow

        Returns
        -------
        Workflow
            Deepcopy
        """
        wf_copy = copy.deepcopy(self)
        return wf_copy

    def plot_dask(self, filename):
        """Save a visualization of workflow to file

        Parameters
        ----------
        filename : str
            Path
        """
        from dask import visualize

        visualize(self.as_dask_dict(), filename=filename, collapse_outputs=True)

    @property
    def tasks(self):
        """All tasks in workflow"""
        return list(self._g.nodes())

    def __len__(self):
        return len(self._g.nodes)

    def __str__(self):
        return '\n'.join([str(task) for task in self._g])

    def __add__(self, other):
        wf_new = Workflow()
        wf_new._g = nx.compose(self._g, other._g)
        return wf_new


class Task:
    """One task

    Parameters
    ----------
    name : str
        Name of task
    function : func
        Task function
    task_input : any
        Input arguments to func
    """

    def __init__(self, name, function, *task_input):
        self.name = name
        self.function = function
        self.task_input = task_input

    def has_input(self):
        """Check if task has any input

        Returns
        -------
        bool
            True if input else False
        """
        return len(self.task_input) > 0

    def __repr__(self):
        return self.name



from pharmpy.workflows import Task, Workflow, local_dask


def fun(s):
    return s


def test_local_dispatcher():
    wf = Workflow([Task('results', fun, 'input')])
    res = local_dask.run(wf)
    assert res == 'input'



import pytest

from pharmpy.estimation import EstimationStep


def test_estimation_method():
    a = EstimationStep('foce', cov=True)
    assert a.method == 'FOCE'
    assert a.cov

    with pytest.raises(ValueError):
        EstimationStep('sklarfs')

    a.method = 'fo'
    assert a.method == 'FO'

    assert a == EstimationStep('fo', interaction=False, cov=True)
    assert (
        repr(a) == 'EstimationStep("FO", interaction=False, cov=True, evaluation=False, '
        'maximum_evaluations=None, laplace=False, isample=None, niter=None, auto=None, '
        'keep_every_nth_iter=None, tool_options={})'
    )

    with pytest.raises(ValueError):
        EstimationStep('fo', maximum_evaluations=0)


def test_append_options():
    a = EstimationStep('foce')

    a.tool_options.update({'EONLY': 1})
    assert len(a.tool_options) == 1

    a.tool_options.update({'PRINT': 1})
    assert len(a.tool_options) == 2


def test_copy(datadir):
    a = EstimationStep('foce', cov=True)
    b = a.copy()
    assert id(a) != id(b)
    assert b.method == 'FOCE'



from pharmpy.results import Results
from pharmpy.workflows import Task, Workflow


def create_workflow(model):
    wf = Workflow()
    wf.name = 'evaldesign'

    from pharmpy.plugins.nonmem.run import evaluate_design

    task = Task('run', evaluate_design, model)
    wf.add_task(task)

    task_result = Task('results', post_process_results)
    wf.add_task(task_result, predecessors=[task])
    return wf


def post_process_results(res):
    return res


class EvalDesignResults(Results):
    def __init__(self, ofv=None, individual_ofv=None, information_matrix=None):
        self.ofv = ofv
        self.individual_ofv = individual_ofv
        self.information_matrix = information_matrix



import os
import subprocess
import uuid
from pathlib import Path

from pharmpy.modeling import write_csv, write_model
from pharmpy.plugins.nonmem import conf, convert_model


def execute_model(model):
    database = model.database
    parent_model = model.parent_model
    model = convert_model(model)
    path = Path.cwd() / f'NONMEM_run_{model.name}-{uuid.uuid1()}'
    path.mkdir(parents=True, exist_ok=True)
    model = model.copy()
    model.parent_model = parent_model
    write_csv(model, path=path, force=True)
    # Set local path
    model.datainfo.path = model.datainfo.path.relative_to(model.datainfo.path.parent)
    model._dataset_updated = True  # Hack to get update_source to update IGNORE
    write_model(model, path=path, force=True)
    basepath = Path(model.name)
    args = [
        nmfe_path(),
        model.name + model.filename_extension,
        str(basepath.with_suffix('.lst')),
    ]
    # Create wrapper script that cd:s into rundirectory
    # This enables the execute_model function to be parallelized
    # using threads. chdir here does not work since all threads will
    # share cwd. Using processes on Windows from R currently hangs.
    # Also the -rundir option to nmfe does not work entirely
    if os.name == 'nt':
        with open(path / 'cdwrapper.bat', 'w') as fp:
            fp.write(f"@echo off\ncd {path}\n{' '.join(args)}\n")
        cmd = str(path / 'cdwrapper.bat')
    else:
        with open(path / 'cdwrapper', 'w') as fp:
            fp.write(f"#!/bin/sh\ncd {path}\n{' '.join(args)}\n")
        os.chmod(path / 'cdwrapper', 0o744)
        cmd = str(path / 'cdwrapper')
    subprocess.call(
        [cmd], stdin=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
    )
    database.store_local_file(model, (path / basepath).with_suffix(model.filename_extension))
    database.store_local_file(model, (path / basepath).with_suffix('.lst'))
    database.store_local_file(model, (path / basepath).with_suffix('.ext'))
    database.store_local_file(model, (path / basepath).with_suffix('.phi'))
    database.store_local_file(model, (path / basepath).with_suffix('.cov'))
    database.store_local_file(model, (path / basepath).with_suffix('.cor'))
    database.store_local_file(model, (path / basepath).with_suffix('.coi'))
    if False:
        # FIXME: Have option for saving the dataset.
        database.store_local_file(model, model.datainfo.path)
    for rec in model.control_stream.get_records('TABLE'):
        database.store_local_file(model, path / rec.path)
    # Read in results for the server side
    model.read_modelfit_results()
    # FIXME: the database path is changed in write
    model.database = database

    return model


def nmfe_path():
    if os.name == 'nt':
        nmfe_candidates = ['nmfe74.bat', 'nmfe75.bat', 'nmfe73.bat']
    else:
        nmfe_candidates = ['nmfe74', 'nmfe75', 'nmfe73']
    path = conf.default_nonmem_path
    if path != Path(''):
        path /= 'run'
    for nmfe in nmfe_candidates:
        candidate_path = path / nmfe
        if candidate_path.is_file():
            path = candidate_path
            break
    else:
        raise FileNotFoundError(f'Cannot find nmfe script for NONMEM ({path})')
    return str(path)


def evaluate_design(model):
    # Prepare and run model for design evaluation
    model = model.copy()
    model.name = '_design_model'

    estrecs = model.control_stream.get_records('ESTIMATION')
    model.control_stream.remove_records(estrecs)

    design_code = '$DESIGN APPROX=FOCEI MODE=1 NELDER FIMDIAG=0 DATASIM=1 GROUPSIZE=32 OFVTYPE=0'
    model.control_stream.insert_record(design_code)

    execute_model(model)

    from pharmpy.tools.evaldesign import EvalDesignResults

    res = EvalDesignResults(
        ofv=model.modelfit_results.ofv,
        individual_ofv=model.modelfit_results.individual_ofv,
        information_matrix=model.modelfit_results.information_matrix,
    )
    return res
